from circuits.web import Controller
from circuits.web.tools import basic_auth, check_auth

from .helpers import HTTPBasicAuthHandler, HTTPError, build_opener, install_opener, urlopen


class Root(Controller):
    def index(self):
        realm = 'Test'
        users = {'admin': 'admin'}
        encrypt = str

        if check_auth(self.request, self.response, realm, users, encrypt):
            return 'Hello World!'

        return basic_auth(self.request, self.response, realm, users, encrypt)


def test(webapp):
    try:
        f = urlopen(webapp.server.http.base)
    except HTTPError as e:
        assert e.code == 401
        assert e.msg == 'Unauthorized'
    else:
        assert False

    handler = HTTPBasicAuthHandler()
    handler.add_password('Test', webapp.server.http.base, 'admin', 'admin')
    opener = build_opener(handler)
    install_opener(opener)

    f = urlopen(webapp.server.http.base)
    s = f.read()
    assert s == b'Hello World!'
    install_opener(None)



"""Help Tasks"""

from fabric import state
from fabric.api import task
from fabric.task_utils import crawl
from fabric.tasks import Task


@task(default=True)
def help(name=None):
    """
    Display help for a given task

    Options:
        name    - The task to display help on.

    To display a list of available tasks type:

        $ fab -l

    To display help on a specific task type:

        $ fab help:<name>
    """
    if name is None:
        name = 'help'

    task = crawl(name, state.commands)
    if isinstance(task, Task):
        doc = getattr(task, '__doc__', None)
        if doc is not None:
            print(f'Help on {name:s}:')
            print()
            print(doc)
        else:
            print(f'No help available for {name:s}')
    else:
        print(f'No such task {name:s}')
        print('For a list of tasks type: fab -l')



""".. codeauthor: mnl"""

import os
import random

from circuits.core.components import BaseComponent
from circuits.core.handlers import handler
from circuits.net.events import close, read, write


class WebSocketCodec(BaseComponent):
    """
    WebSocket Protocol

    Implements the Data Framing protocol for WebSocket.

    This component is used in conjunction with a parent component that
    receives Read events on its channel. When created (after a successful
    WebSocket setup handshake), the codec registers
    a handler on the parent's channel that filters out these Read
    events for a given socket (if used in a server) or all Read events
    (if used in a client). The data is decoded and the contained payload
    is emitted as Read events on the codec's channel.

    The data from write events sent to the codec's channel
    (with socket argument if used in a server) is
    encoded according to the WebSocket Data Framing protocol. The
    encoded data is then forwarded as write events on the parents channel.
    """

    channel = 'ws'

    def __init__(self, sock=None, data=bytearray(), *args, **kwargs):
        """
        Creates a new codec.

        :param sock: the socket used in Read and write events
            (if used in a server, else None)
        """
        super().__init__(*args, **kwargs)

        self._sock = sock

        self._pending_payload = bytearray()
        self._pending_type = None
        self._close_received = False
        self._close_sent = False
        self._buffer = bytearray()

        messages = self._parse_messages(bytearray(data))
        for message in messages:
            if self._sock is not None:
                self.fire(read(self._sock, message))
            else:
                self.fire(read(message))

    @handler('registered')
    def _on_registered(self, component, parent):
        if component is self:

            @handler('read', priority=10, channel=parent.channel)
            def _on_read_raw(self, event, *args):
                if self._sock is not None:
                    if args[0] != self._sock:
                        return
                    data = args[1]
                else:
                    data = args[0]
                messages = self._parse_messages(bytearray(data))
                for message in messages:
                    if self._sock is not None:
                        self.fire(read(self._sock, message))
                    else:
                        self.fire(read(message))
                event.stop()

            self.addHandler(_on_read_raw)

            @handler('disconnect', channel=parent.channel)
            def _on_disconnect(self, *args):
                if self._sock is not None and args[0] != self._sock:
                    return
                self.unregister()

            self.addHandler(_on_disconnect)

    def _parse_messages(self, data):
        msgs = []  # one chunk of bytes may result in several messages
        if self._close_received:
            return msgs
        data = self._buffer + data
        while data:
            # extract final flag, opcode and masking
            final = bool(data[0] & 0x80 != 0)
            opcode = data[0] & 0xF
            masking = bool(data[1] & 0x80 != 0)
            # evaluate payload length
            payload_length = data[1] & 0x7F
            offset = 2
            if payload_length >= 126:
                payload_bytes = 2 if payload_length == 126 else 8
                payload_length = 0
                for _ in range(payload_bytes):
                    payload_length = payload_length * 256 + data[offset]
                    offset += 1
            # retrieve optional masking key
            if masking:
                masking_key = data[offset : offset + 4]
                offset += 4
            # if not enough bytes available yet, retry after next read
            if len(data) - offset < payload_length:
                self._buffer = data
                break
            self._buffer = bytearray()
            # rest of _buffer is payload
            msg = data[offset : offset + payload_length]
            if masking:  # unmask
                msg = list(msg)
                for i, c in enumerate(msg):
                    msg[i] = c ^ masking_key[i % 4]
                msg = bytearray(msg)
            # remove bytes of processed frame from byte _buffer
            offset += payload_length
            data = data[offset:]
            # if there have been parts already, combine
            msg = self._pending_payload + msg
            if final:
                if opcode < 8:
                    # if text or continuation of text, convert
                    if opcode == 1 or (opcode == 0 and self._pending_type == 1):
                        msg = msg.decode('utf-8', 'replace')
                    self._pending_type = None
                    self._pending_payload = bytearray()
                    msgs.append(msg)
                # check for client closing the connection
                elif opcode == 8:
                    self._close_received = True
                    if self._sock:
                        self.fire(close(self._sock))
                    else:
                        self.fire(close())
                    break
                # check for Ping
                elif opcode == 9:
                    if self._close_sent:
                        return None
                    frame = bytearray(b'\x8a')
                    frame += self._encode_tail(msg, self._sock is None)
                    self._write(frame)
            else:
                self._pending_payload = msg
                if opcode != 0:
                    self._pending_type = opcode
        return msgs

    @handler('write')
    def _on_write(self, *args):
        if self._close_sent:
            return

        if self._sock is not None:
            if args[0] != self._sock:
                return
            data = args[1]
        else:
            data = args[0]

        frame = bytearray()
        first = 0x80  # set FIN flag, we never fragment
        if isinstance(data, str):
            first += 1  # text
            data = bytearray(data, 'utf-8')
        else:
            first += 2  # binary
        frame.append(first)
        frame += self._encode_tail(data, self._sock is None)
        self._write(frame)

    def _encode_tail(self, data, mask=False):
        tail = bytearray()
        data_length = len(data)
        if data_length <= 125:
            len_byte = data_length
            lbytes = 0
        elif data_length <= 0xFFFF:
            len_byte = 126
            lbytes = 2
        else:
            len_byte = 127
            lbytes = 8
        if mask:
            len_byte = len_byte | 0x80
        tail.append(len_byte)
        for i in range(lbytes - 1, -1, -1):
            tail.append(data_length >> (i * 8) & 0xFF)
        if mask:
            try:
                masking_key = bytearray(list(os.urandom(4)))
            except NotImplementedError:
                masking_key = bytearray(random.randint(0, 255) for i in range(4))
            tail += masking_key
            for i, c in enumerate(data):
                tail.append(c ^ masking_key[i % 4])
        else:
            tail += data
        return tail

    def _write(self, data):
        if self._sock is not None:
            self.fire(write(self._sock, data), self.parent.channel)
        else:
            self.fire(write(data), self.parent.channel)

    @handler('close')
    def _on_close(self, *args):
        if self._sock is not None and args and (args[0] != self._sock):
            return
        if not self._close_sent:
            self._write(b'\x88\x00')
            self._close_sent = True
        if self._close_received and self._close_sent:
            if self._sock:
                self.fire(close(self._sock), self.parent.channel)
            else:
                self.fire(close(), self.parent.channel)



#!/usr/bin/env python
from circuits.web import Controller

from .helpers import urlopen


class Root(Controller):
    def index(self):
        return 'Hello World!'

    def foo(self):
        self.response.headers['Content-Type'] = 'text/plain'
        return 'Hello World!'

    def empty(self):
        return ''


def test_default(webapp):
    f = urlopen(webapp.server.http.base)
    s = f.read()
    assert s == b'Hello World!'

    content_type = f.headers['Content-Type']
    assert content_type == 'text/html; charset=utf-8'


def test_explicit(webapp):
    f = urlopen(f'{webapp.server.http.base:s}/foo')
    s = f.read()
    assert s == b'Hello World!'

    content_type = f.headers['Content-Type']
    assert content_type == 'text/plain'


def test_static(webapp):
    f = urlopen(f'{webapp.server.http.base:s}/static/test.css')
    s = f.read()
    assert s == b'body { }\n'

    content_type = f.headers['Content-Type']
    assert content_type == 'text/css'


def test_empty(webapp):
    f = urlopen(f'{webapp.server.http.base:s}/empty')
    s = f.read()
    assert s == b''

    content_length = f.headers['Content-Length']
    assert int(content_length) == 0



"""Workers Tests"""

from os import getpid

import pytest

from circuits import Worker, task


@pytest.fixture()
def worker(request, manager):
    worker = Worker().register(manager)

    def finalizer():
        worker.unregister()

    request.addfinalizer(finalizer)

    return worker


def err():
    return x * 2  # noqa: F821


def foo():
    x = 0
    i = 0
    while i < 1000000:
        x += 1
        i += 1
    return x


def pid():
    return f'Hello from {getpid():d}'


def add(a, b):
    return a + b


def test_failure(manager, watcher, worker):
    e = task(err)
    e.failure = True

    x = worker.fire(e)

    assert watcher.wait('task_failure')

    assert isinstance(x.value[1], Exception)


def test_success(manager, watcher, worker):
    e = task(foo)
    e.success = True

    x = worker.fire(e)

    assert watcher.wait('task_success')

    assert x.value == 1000000


def test_args(manager, watcher, worker):
    e = task(add, 1, 2)
    e.success = True

    x = worker.fire(e)

    assert watcher.wait('task_success')

    assert x.value == 3



#!/usr/bin/env python
from socket import gaierror


def test_client_bind_int():
    from circuits.net.sockets import Client

    class TestClient(Client):
        def _create_socket(self):
            return None

    client = TestClient(1234)

    assert client._bind[1] == 1234


def test_client_bind_int_gaierror(monkeypatch):
    from circuits.net import sockets

    def broken_gethostname():
        raise gaierror()

    monkeypatch.setattr(sockets, 'gethostname', broken_gethostname)

    class TestClient(sockets.Client):
        def _create_socket(self):
            return None

    client = TestClient(1234)

    assert client._bind == ('0.0.0.0', 1234)


def test_client_bind_str():
    from circuits.net.sockets import Client

    class TestClient(Client):
        def _create_socket(self):
            return None

    client = TestClient('0.0.0.0:1234')

    assert client._bind == ('0.0.0.0', 1234)



#!/usr/bin/env python
from circuits import Debugger
from circuits.web import Controller, Server, Static


class Root(Controller):
    def index(self):
        """
        Index Request Handler

        Controller(s) expose implicitly methods as request handlers.
        Request Handlers can still be customized by using the ``@expose``
        decorator. For example exposing as a different path.
        """
        return 'Hello World!'


app = Server(('0.0.0.0', 8000))
Debugger().register(app)
Static().register(app)
Root().register(app)
app.run()



#!/usr/bin/env python
from circuits.core.components import BaseComponent
from circuits.core.handlers import handler
from circuits.web import BaseServer, Controller
from circuits.web.dispatchers.dispatcher import Dispatcher

from .helpers import urljoin, urlopen


class PrefixingDispatcher(BaseComponent):
    """Forward to another Dispatcher based on the channel."""

    @handler('request', priority=1.0)
    def _on_request(self, event, request, response):
        path = request.path.strip('/')

        path = urljoin('/%s/' % self.channel, path)
        request.path = path


class DummyRoot(Controller):
    channel = '/'

    def index(self):
        return 'Not used'


class Root1(Controller):
    channel = '/site1'

    def index(self):
        return 'Hello from site 1!'


class Root2(Controller):
    channel = '/site2'

    def index(self):
        return 'Hello from site 2!'


def test_disps(manager, watcher):
    server1 = BaseServer(0, channel='site1').register(manager)
    PrefixingDispatcher(channel='site1').register(manager)
    Dispatcher(channel='site1').register(manager)
    Root1().register(manager)
    assert watcher.wait('ready', channel=server1.channel)

    server2 = BaseServer(('localhost', 0), channel='site2').register(manager)
    PrefixingDispatcher(channel='site2').register(manager)
    Dispatcher(channel='site2').register(manager)
    Root2().register(manager)
    assert watcher.wait('ready', channel=server2.channel)

    DummyRoot().register(manager)

    f = urlopen(server1.http.base, timeout=3)
    s = f.read()
    assert s == b'Hello from site 1!'

    f = urlopen(server2.http.base, timeout=3)
    s = f.read()
    assert s == b'Hello from site 2!'



#!/usr/bin/env python
import pytest

from circuits.io import Process, write


if pytest.PLATFORM == 'win32':
    pytest.skip('Unsupported Platform')


def test(manager, watcher):
    p = Process(['echo', 'Hello World!']).register(manager)
    assert watcher.wait('registered')

    p.start()
    assert watcher.wait('started', p.channel)

    assert watcher.wait('terminated', p.channel)

    s = p.stdout.getvalue()
    assert s == b'Hello World!\n'


def test2(manager, watcher, tmpdir):
    foo = tmpdir.ensure('foo.txt')

    p = Process([f'cat - > {foo!s:s}'], shell=True).register(manager)
    assert watcher.wait('registered')

    p.start()
    assert watcher.wait('started', p.channel)

    p.fire(write('Hello World!'), p._stdin)
    assert watcher.wait('write', p._stdin)

    p.stop()

    assert watcher.wait('eof', p._stdout.channel)

    with foo.open('r') as f:
        assert f.read() == 'Hello World!'


def test_two_procs(manager, watcher):
    p1 = Process(['echo', '1']).register(manager)
    p2 = Process('echo 2 ; sleep 1', shell=True).register(manager)

    p1.start()
    p2.start()

    assert watcher.wait('terminated', p1.channel)
    assert p1._terminated
    assert not p2._terminated
    assert not p2._stdout_closed
    assert not p2._stderr_closed

    watcher.clear()  # Get rid of first terminated()

    s1 = p1.stdout.getvalue()
    assert s1 == b'1\n'

    assert watcher.wait('terminated', p2.channel)
    assert p2._terminated
    assert p2._stdout_closed
    assert p2._stderr_closed

    s2 = p2.stdout.getvalue()
    assert s2 == b'2\n'



#!/usr/bin/env python
from circuits import Component
from circuits.web import BaseServer


class Root(Component):
    def request(self, request, response):
        """
        Request Handler

        Using ``BaseServer`` provides no URL dispatching of any kind.
        Requests are handled by any event handler listening to
        ``Request`` events and must accept a (request, response)
        as arguments.
        """
        return 'Hello World!'


app = BaseServer(('0.0.0.0', 8000))
Root().register(app)
app.run()



"""Timers Tests"""

from datetime import datetime, timedelta
from itertools import starmap
from operator import sub
from time import time

import pytest

from circuits import Component, Event, Timer, sleep


@pytest.fixture()
def app(request, manager, watcher):
    app = App().register(manager)
    assert watcher.wait('registered')

    def finalizer():
        app.unregister()
        assert watcher.wait('unregistered')

    request.addfinalizer(finalizer)

    return app


class single(Event):
    """single Event"""

    complete = True


class persistent(Event):
    """persistent Event"""

    complete = True


class App(Component):
    def init(self):
        self.flag = False
        self.count = 0
        self.timestamps = []

    def single(self):
        self.timestamps.append(time())
        self.count += 1
        self.flag = True

    def persistent(self, interval):
        timer = Timer(interval, single(), persist=True)
        timer.register(self)

        yield sleep(interval * 10)

        timer.unregister()


def test_single(app, watcher):
    Timer(0.1, single()).register(app)
    assert watcher.wait('single_complete')
    assert app.flag


def test_persistent(app, watcher):
    exponent = -1
    interval = 10.0**exponent
    app.fire(persistent(interval))
    assert watcher.wait('persistent_complete')

    xs = list(map(abs, starmap(sub, zip(app.timestamps, app.timestamps[1:]))))
    avg = sum(xs) / len(xs)

    assert round(avg, abs(exponent)) == interval


def test_datetime(app, watcher):
    now = datetime.now()
    d = now + timedelta(seconds=0.1)
    Timer(d, single()).register(app)
    assert watcher.wait('single_complete')
    assert app.flag



#!/usr/bin/env python
from json import dumps

from circuits import Component, handler
from circuits.web import Controller, Logger, Server


class JSONSerializer(Component):
    channel = 'web'

    # 1 higher than the default response handler
    @handler('response', priority=1.0)
    def serialize_response_body(self, response):
        response.headers['Content-Type'] = 'application/json'
        response.body = dumps(response.body)


class Root(Controller):
    def index(self):
        return {'message': 'Hello World!'}


app = Server(('0.0.0.0', 8000))
JSONSerializer().register(app)
Logger().register(app)
Root().register(app)
app.run()



from http.cookiejar import CookieJar  # noqa
from urllib.error import HTTPError, URLError  # noqa
from urllib.parse import quote, urlencode, urljoin, urlparse  # noqa
from urllib.request import (  # noqa
    HTTPBasicAuthHandler,
    HTTPCookieProcessor,
    HTTPDigestAuthHandler,
    Request,
    build_opener,
    install_opener,
    urlopen,
)


# pylama:skip=1



#!/usr/bin/env python
import pytest
from pytest import fixture

from circuits import Component, Event, handler
from circuits.net.events import read, write
from circuits.protocols.irc import (
    AWAY,
    INVITE,
    IRC,
    JOIN,
    KICK,
    MODE,
    NAMES,
    NICK,
    NOTICE,
    PART,
    PASS,
    PONG,
    PRIVMSG,
    QUIT,
    TOPIC,
    USER,
    WHOIS,
    irc_color_to_ansi,
    joinprefix,
    parsemsg,
    parseprefix,
    strip,
)


class App(Component):
    def init(self):
        IRC().register(self)

        self.data = []
        self.events = []

    @handler(False)
    def reset(self):
        self.data = []
        self.events = []

    @handler()
    def _on_event(self, event, *args, **kwargs):
        self.events.append(event)

    def request(self, message):
        self.fire(write(bytes(message)))

    def write(self, data):
        self.data.append(data)


@fixture()
def app(request):
    app = App()

    while len(app):
        app.flush()

    return app


def test_strip():
    s = ':\x01\x02test\x02\x01'
    s = strip(s)
    assert s == '\x01\x02test\x02\x01'

    s = ':\x01\x02test\x02\x01'
    s = strip(s, color=True)
    assert s == 'test'


def test_joinprefix():
    nick, ident, host = 'test', 'foo', 'localhost'
    s = joinprefix(nick, ident, host)
    assert s == 'test!foo@localhost'


def test_parsemsg():
    s = b':foo!bar@localhost NICK foobar'
    source, command, args = parsemsg(s)
    assert source == ('foo', 'bar', 'localhost')
    assert command == 'NICK'
    assert args == ['foobar']

    s = b''
    source, command, args = parsemsg(s)
    assert source == (None, None, None)
    assert command is None
    assert args == []


def test_parseprefix():
    s = 'test!foo@localhost'
    nick, ident, host = parseprefix(s)
    assert nick == 'test'
    assert ident == 'foo'
    assert host == 'localhost'

    s = 'test'
    nick, ident, host = parseprefix(s)
    assert nick == 'test'
    assert ident is None
    assert host is None


@pytest.mark.parametrize(
    'event,data',
    [
        (PASS('secret'), b'PASS secret\r\n'),
        (
            USER('foo', 'localhost', 'localhost', 'Test Client'),
            b'USER foo localhost localhost :Test Client\r\n',
        ),
        (NICK('test'), b'NICK test\r\n'),
        pytest.param(PONG('localhost'), b'PONG :localhost\r\n', marks=pytest.mark.xfail),
        pytest.param(QUIT(), b'QUIT Leaving\r\n', marks=pytest.mark.xfail),
        (QUIT('Test'), b'QUIT Test\r\n'),
        (QUIT('Test Message'), b'QUIT :Test Message\r\n'),
        (JOIN('#test'), b'JOIN #test\r\n'),
        (JOIN('#test', 'secret'), b'JOIN #test secret\r\n'),
        (PART('#test'), b'PART #test\r\n'),
        (PRIVMSG('test', 'Hello'), b'PRIVMSG test Hello\r\n'),
        (PRIVMSG('test', 'Hello World'), b'PRIVMSG test :Hello World\r\n'),
        (NOTICE('test', 'Hello'), b'NOTICE test Hello\r\n'),
        (NOTICE('test', 'Hello World'), b'NOTICE test :Hello World\r\n'),
        pytest.param(KICK('#test', 'test'), b'KICK #test test :\r\n', marks=pytest.mark.xfail),
        (KICK('#test', 'test', 'Bye'), b'KICK #test test Bye\r\n'),
        (KICK('#test', 'test', 'Good Bye!'), b'KICK #test test :Good Bye!\r\n'),
        (TOPIC('#test', 'Hello World!'), b'TOPIC #test :Hello World!\r\n'),
        (MODE('+i'), b'MODE +i\r\n'),
        (MODE('#test', '+o', 'test'), b'MODE #test +o test\r\n'),
        (INVITE('test', '#test'), b'INVITE test #test\r\n'),
        pytest.param(NAMES(), b'NAMES\r\n', marks=pytest.mark.xfail),
        (NAMES('#test'), b'NAMES #test\r\n'),
        (AWAY('I am away.'), b'AWAY :I am away.\r\n'),
        pytest.param(WHOIS('somenick'), b'WHOIS :somenick\r\n', marks=pytest.mark.xfail),
    ],
)
def test_commands(event, data):
    message = event.args[0]
    assert bytes(message) == data


@pytest.mark.parametrize(
    'data,event',
    [
        (
            b':localhost NOTICE * :*** Looking up your hostname...\r\n',
            Event.create(
                'notice',
                ('localhost', None, None),
                '*',
                '*** Looking up your hostname...',
            ),
        ),
    ],
)
def test_responses(app, data, event):
    app.reset()
    app.fire(read(data))
    while len(app):
        app.flush()

    e = app.events[-1]

    assert event.name == e.name
    assert event.args == e.args
    assert event.kwargs == e.kwargs


@pytest.mark.parametrize(
    'inp,out',
    [
        (
            'hi \x02bold\x02 \x1ditalic\x1d \x1funderline\x1f \x1estrikethrough\x1e',
            'hi \x1b[01mbold\x1b[22m \x1b[03mitalic\x1b[23m \x1b[04munderline\x1b[24m \x1b[09mstrikethrough\x1b[29m',
        ),  # noqa: E501
        (
            '\x0300white\x03 \x0301black\x03 \x0302blue\x03 \x0303green\x03 \x0304red\x03 ',
            '\x1b[37mwhite\x1b[39;49m \x1b[30mblack\x1b[39;49m \x1b[34mblue\x1b[39;49m \x1b[32mgreen\x1b[39;49m \x1b[31mred\x1b[39;49m ',
        ),  # noqa: E501
        (
            '\x0305brown\x03 \x0306magenta\x03 \x0307orange\x03 \x0308yellow\x03 ',
            '\x1b[36mbrown\x1b[39;49m \x1b[35mmagenta\x1b[39;49m \x1b[33morange\x1b[39;49m \x1b[93myellow\x1b[39;49m ',
        ),  # noqa: E501
        (
            '\x0309lightgreen\x03 \x0310cyan\x03 \x0311lightcyan\x03 \x0312lightblue\x03 ',
            '\x1b[92mlightgreen\x1b[39;49m \x1b[36mcyan\x1b[39;49m \x1b[96mlightcyan\x1b[39;49m \x1b[94mlightblue\x1b[39;49m ',
        ),  # noqa: E501
        (
            '\x0313pink\x03 \x0314grey\x03 \x0315lightgrey\x03',
            '\x1b[95mpink\x1b[39;49m \x1b[90mgrey\x1b[39;49m \x1b[37mlightgrey\x1b[39;49m',
        ),  # noqa: E501
        (
            '\x0300white\x03 \x0301,01black\x03 \x0301,02blue\x03 \x0301,03green\x03 \x0301,04red\x03 ',
            '\x1b[37mwhite\x1b[39;49m \x1b[30;40mblack\x1b[39;49m \x1b[30;44mblue\x1b[39;49m \x1b[30;42mgreen\x1b[39;49m \x1b[30;41mred\x1b[39;49m ',
        ),  # noqa: E501
        ('\x0f', '\x1b[m'),
        ('\x0302blue', '\x1b[34mblue\x1b[m'),
    ],
)
def test_ansi_color(inp, out):
    assert irc_color_to_ansi(inp) == out



#!/usr/bin/env python
import pytest

from circuits import Manager
from circuits.core.pollers import Select
from circuits.net.events import close, write
from circuits.net.sockets import Pipe

from .client import Client


pytestmark = pytest.mark.skipif(pytest.PLATFORM == 'win32', reason='Unsupported Platform')


def pytest_generate_tests(metafunc):
    metafunc.parametrize('Poller', [Select])


def test_pipe(Poller):
    m = Manager() + Poller()

    a, b = Pipe('a', 'b')
    a.register(m)
    b.register(m)

    a = Client(channel=a.channel).register(m)
    b = Client(channel=b.channel).register(m)

    m.start()

    try:
        assert pytest.wait_for(a, 'ready')
        assert pytest.wait_for(b, 'ready')

        a.fire(write(b'foo'))
        assert pytest.wait_for(b, 'data', b'foo')

        b.fire(write(b'foo'))
        assert pytest.wait_for(a, 'data', b'foo')

        a.fire(close())
        assert pytest.wait_for(a, 'disconnected')

        b.fire(close())
        assert pytest.wait_for(b, 'disconnected')
    finally:
        m.stop()



"""I/O Support

This package contains various I/O Components. Provided are a generic File
Component, StdIn, StdOut and StdErr components. Instances of StdIn, StdOut
and StdErr are also created by importing this package.
"""

import sys

from .events import close, open, seek, write
from .file import File
from .process import Process


try:
    from .notify import Notify
except:
    pass

try:
    from .serial import Serial
except:
    pass

try:
    stdin = File(sys.stdin, channel='stdin')
    stdout = File(sys.stdout, channel='stdout')
    stderr = File(sys.stderr, channel='stderr')
except:
    pass

# flake8: noqa
# pylama: skip=1



"""
JSON RPC

This module implements a JSON RPC dispatcher that translates incoming
RPC calls over JSON into RPC events.
"""

import json

from circuits import BaseComponent, Event, handler


class rpc(Event):
    """RPC Event"""


class JSONRPC(BaseComponent):
    channel = 'web'

    def __init__(self, path=None, encoding='utf-8', rpc_channel='*'):
        super().__init__()

        if json is None:
            raise RuntimeError('No json support available')

        self.path = path
        self.encoding = encoding
        self.rpc_channel = rpc_channel

    @handler('request', priority=0.2)
    def _on_request(self, event, req, res):
        if self.path is not None and self.path != req.path.rstrip('/'):
            return

        res.headers['Content-Type'] = 'application/json'

        try:
            data = req.body.read().decode(self.encoding)
            o = json.loads(data)
            id, method, params = o['id'], o['method'], o.get('params', {})
            if isinstance(params, dict):
                params = {str(k): v for k, v in params.items()}

            method = str(method)

            if isinstance(params, dict):
                value = yield self.call(rpc.create(method, **params), self.rpc_channel)
            else:
                value = yield self.call(rpc.create(method, *params), self.rpc_channel)

            yield self._response(id, value.value)
        except Exception as e:
            yield self._error(-1, 100, f'{e.__class__.__name__}: {e}')
        finally:
            event.stop()

    def _response(self, id, result):
        data = {
            'id': id,
            'version': '1.1',
            'result': result,
            'error': None,
        }
        return json.dumps(data).encode(self.encoding)

    def _error(self, id, code, message):
        data = {
            'id': id,
            'version': '1.1',
            'error': {
                'name': 'JSONRPCError',
                'code': code,
                'message': message,
            },
        }
        return json.dumps(data).encode(self.encoding)



#!/usr/bin/env python
import json

from circuits.web import Controller
from circuits.web.exceptions import Forbidden, NotFound, Redirect

from .helpers import HTTPError, urlopen


class Root(Controller):
    def index(self):
        return 'Hello World!'

    def test_redirect(self):
        raise Redirect('/')

    def test_forbidden(self):
        raise Forbidden()

    def test_notfound(self):
        raise NotFound()

    def test_contenttype(self):
        raise Exception()

    def test_contenttype_json(self):
        self.response.headers['Content-Type'] = 'application/json'
        raise Exception()

    def test_contenttype_json_no_debug(self):
        self.response.headers['Content-Type'] = 'application/json'
        self.request.print_debug = False
        raise Exception()


def test_redirect(webapp):
    f = urlopen('%s/test_redirect' % webapp.server.http.base)
    s = f.read()
    assert s == b'Hello World!'


def test_forbidden(webapp):
    try:
        urlopen('%s/test_forbidden' % webapp.server.http.base)
    except HTTPError as e:
        assert e.code == 403
        assert e.msg == 'Forbidden'
    else:
        assert False


def test_notfound(webapp):
    try:
        urlopen('%s/test_notfound' % webapp.server.http.base)
    except HTTPError as e:
        assert e.code == 404
        assert e.msg == 'Not Found'
    else:
        assert False


def test_contenttype(webapp):
    try:
        urlopen('%s/test_contenttype' % webapp.server.http.base)
    except HTTPError as e:
        assert e.code == 500
        assert e.msg == 'Internal Server Error'
        assert 'text/html' in e.headers.get('Content-Type')
    else:
        assert False


def test_contenttype_json(webapp):
    try:
        urlopen('%s/test_contenttype_json' % webapp.server.http.base)
    except HTTPError as e:
        assert 'json' in e.headers.get('Content-Type')
        result = json.loads(e.read().decode('utf-8'))
        assert result['code'] == 500
        assert result['name'] == 'Internal Server Error'
        assert result['description'] == ''
        assert 'raise Exception' in result['traceback']
    else:
        assert False


def test_contenttype_json_no_debug(webapp):
    try:
        urlopen('%s/test_contenttype_json_no_debug' % webapp.server.http.base)
    except HTTPError as e:
        assert 'json' in e.headers.get('Content-Type')
        result = json.loads(e.read().decode('utf-8'))
        assert result['code'] == 500
        assert result['name'] == 'Internal Server Error'
        assert result['description'] == ''
        assert 'traceback' not in result
    else:
        assert False



"""
HTML macros

Macros for generating snippets of HTML.
"""

import genshi
import pygments
import pygments.formatters
import pygments.lexers
import pygments.util
from genshi import builder
from genshi.filters import HTMLSanitizer


sanitizer = HTMLSanitizer()


def pre(macro, environ, *args, **kwargs):
    """
    Return the raw text of body, rendered in a <pre> block.

    **Arguments:** //None//

    **Example:**
    {{{
    <<pre>>
    def hello():
        print "Hello World!"

    hello()
    <</pre>>
    }}}
    """
    if macro.body is None:
        return None

    return builder.tag.pre(macro.body)


def code(macro, environ, *args, **kwargs):
    """Render syntax highlighted code"""
    if not macro.body:
        return None

    lang = kwargs.get('lang', None)

    if lang is not None:
        if not macro.isblock:
            return None
        try:
            lexer = pygments.lexers.get_lexer_by_name(lang, stripall=True)
        except pygments.util.ClassNotFound:
            return None
    else:
        lexer = None

    if lexer:
        text = pygments.highlight(
            macro.body,
            lexer,
            pygments.formatters.HtmlFormatter(),
        )
        output = genshi.core.Markup(text)
    elif macro.isblock:
        output = genshi.builder.tag.pre(macro.body)
    else:
        output = genshi.builder.tag.code(
            macro.body,
            style='white-space:pre-wrap',
            class_='highlight',
        )

    return output


def source(macro, environ, *args, **kwargs):
    """Return the parsed text of body, rendered in a <pre> block."""
    if macro.body is None:
        return None

    return builder.tag.pre(environ['parser'].render(macro.body, environ=environ).decode('utf-8'))


def div(macro, environ, cls=None, float=None, id=None, style=None, *args, **kwargs):
    if macro.body is None:
        return None

    if float and float in ('left', 'right'):
        style = f'float: {float}; {style}'

    if style:
        style = ';'.join(sanitizer.sanitize_css(style))

    context = 'block' if macro.isblock else 'inline'

    contents = environ['parser'].generate(
        macro.body,
        environ=environ,
        context=context,
    )

    return builder.tag.div(contents, id=id, class_=cls, style=style)


def span(macro, environ, class_=None, id=None, style=None, *args, **kwargs):
    """..."""
    if macro.body is None:
        return None

    if style:
        style = ';'.join(sanitizer.sanitize_css(style))

    contents = environ['parser'].generate(
        macro.body,
        environ=environ,
        context='inline',
    )

    return builder.tag.span(contents, id=id, class_=class_, style=style)



#!/usr/bin/env python
from circuits.web import Controller, Server, Sessions


class Root(Controller):
    def index(self, name='World'):
        if 'name' in self.session:
            name = self.session['name']
        self.session['name'] = name
        return 'Hello %s!' % name


app = Server(('0.0.0.0', 8000))
Sessions().register(app)
Root().register(app)
app.run()



#!/usr/bin/env python
import os
import ssl
import sys
import tempfile
from os import path
from socket import gaierror

import pytest
from pytest import fixture

from circuits import Component
from circuits.web import BaseServer, Controller, Server

from .helpers import URLError, urlopen


CERTFILE = path.join(path.dirname(__file__), 'cert.pem')


# self signed cert
if pytest.PYVER >= (2, 7, 9):
    SSL_CONTEXT = ssl.SSLContext(ssl.PROTOCOL_TLS)
    SSL_CONTEXT.verify_mode = ssl.CERT_NONE


@fixture()
def tmpfile(request):
    tmpdir = tempfile.mkdtemp()
    return os.path.join(tmpdir, 'test.sock')


class BaseRoot(Component):
    channel = 'web'

    def request(self, request, response):
        return 'Hello World!'


class Root(Controller):
    def index(self):
        return 'Hello World!'


class MakeQuiet(Component):
    channel = 'web'

    def ready(self, event, *args):
        event.stop()


def test_baseserver(manager, watcher):
    server = BaseServer(0).register(manager)
    MakeQuiet().register(server)
    watcher.wait('ready')

    BaseRoot().register(server)
    watcher.wait('registered')

    try:
        f = urlopen(server.http.base)
    except URLError as e:
        if isinstance(e.reason, gaierror):
            f = urlopen('http://127.0.0.1:8000')
        else:
            raise

    s = f.read()
    assert s == b'Hello World!'

    server.unregister()
    watcher.wait('unregistered')


def test_server(manager, watcher):
    server = Server(0).register(manager)
    MakeQuiet().register(server)
    watcher.wait('ready')

    Root().register(server)

    try:
        f = urlopen(server.http.base)
    except URLError as e:
        if isinstance(e.reason, gaierror):
            f = urlopen('http://127.0.0.1:8000')
        else:
            raise

    s = f.read()
    assert s == b'Hello World!'

    server.unregister()
    watcher.wait('unregistered')


@pytest.mark.skipif((2,) < sys.version_info < (3, 4, 3), reason='Context not implemented under python 3.4.3')
@pytest.mark.skipif(sys.version_info < (2, 7, 9), reason='Context not implemented under python 2.7.9')
def test_secure_server(manager, watcher):
    pytest.importorskip('ssl')

    server = Server(0, secure=True, certfile=CERTFILE).register(manager)
    MakeQuiet().register(server)
    watcher.wait('ready')

    Root().register(server)

    try:
        f = urlopen(server.http.base, context=SSL_CONTEXT)
    except URLError as e:
        if isinstance(e.reason, gaierror):
            f = urlopen('http://127.0.0.1:8000')
        else:
            raise

    s = f.read()
    assert s == b'Hello World!'

    server.unregister()
    watcher.wait('unregistered')


def test_unixserver(manager, watcher, tmpfile):
    if pytest.PLATFORM == 'win32':
        pytest.skip('Unsupported Platform')

    server = Server(tmpfile).register(manager)
    MakeQuiet().register(server)
    assert watcher.wait('ready')

    Root().register(server)

    assert path.basename(server.host) == 'test.sock'

    try:
        from uhttplib import UnixHTTPConnection

        client = UnixHTTPConnection(server.http.base)
        client.request('GET', '/')
        response = client.getresponse()
        s = response.read()

        assert s == b'Hello World!'
    except ImportError:
        pass

    server.unregister()
    watcher.wait('unregistered')


@pytest.mark.skipif((2, 7, 9) < sys.version_info < (3, 4, 3), reason='Context not implemented under python 3.4.3')
@pytest.mark.skipif(sys.version_info < (2, 7, 9), reason='Context not implemented under python 2.7.9')
def test_multi_servers(manager, watcher):
    pytest.importorskip('ssl')

    insecure_server = Server(0, channel='insecure')
    secure_server = Server(
        0,
        channel='secure',
        secure=True,
        certfile=CERTFILE,
    )

    server = (insecure_server + secure_server).register(manager)
    MakeQuiet().register(server)
    watcher.wait('ready')

    Root().register(server)

    f = urlopen(insecure_server.http.base)
    s = f.read()
    assert s == b'Hello World!'

    f = urlopen(secure_server.http.base, context=SSL_CONTEXT)
    s = f.read()
    assert s == b'Hello World!'

    server.unregister()
    watcher.wait('unregistered')



#!/usr/bin/env python
"""
Forms

A simple example showing how to deal with data forms.
"""

from circuits.web import Controller, Server


FORM = """
<html>
 <head>
  <title>Basic Form Handling</title>
 </head>
 <body>
  <h1>Basic Form Handling</h1>
  <p>
   Example of using
   <a href="http://circuitsframework.com/">circuits</a> and its
   <b>Web Components</b> to build a simple web application that handles
   some basic form data.
  </p>
  <form action="/save" method="POST">
   <table border="0" rules="none">
    <tr>
     <td>First Name:</td>
     <td><input type="text" name="firstName"></td>
    </tr>
    <tr>
     <td>Last Name:</td>
     <td><input type="text" name="lastName"></td>
    </tr>
     <tr>
      <td colspan=2">
       <input type="submit" value="Save">
     </td>
     </tr>
   </table>
  </form>
 </body>
</html>"""


class Root(Controller):
    def index(self):
        """
        Request Handler

        Our index request handler which simply returns a response containing
        the contents of our form to display.
        """
        return FORM

    def save(self, firstName, lastName):
        """
        Save Request Handler

        Our /save request handler (which our form above points to).
        This handler accepts the same arguments as the fields in the
        form either as positional arguments or keyword arguments.

        We will use the date to pretend we've saved the data and
        tell the user what was saved.
        """
        return f'Data Saved. firstName={firstName:s} lastName={lastName:s}'


app = Server(('0.0.0.0', 8000))
Root().register(app)
app.run()



#!/usr/bin/env python
from io import BytesIO
from os import path

from pytest import fixture

from circuits import Component, handler
from circuits.web import Controller
from circuits.web.tools import gzip

from .conftest import DOCROOT
from .helpers import Request, build_opener


class Gzip(Component):
    channel = 'web'

    @handler('response', priority=1.0)
    def _on_response(self, event, *args, **kwargs):
        event[0] = gzip(event[0])


class Root(Controller):
    def index(self):
        return 'Hello World!'


@fixture()
def gziptool(request, webapp):
    gziptool = Gzip().register(webapp)

    def finalizer():
        gziptool.unregister()

    request.addfinalizer(finalizer)

    return gziptool


def decompress(body):
    import gzip

    zbuf = BytesIO()
    zbuf.write(body)
    zbuf.seek(0)
    zfile = gzip.GzipFile(mode='rb', fileobj=zbuf)
    data = zfile.read()
    zfile.close()
    return data


def test1(webapp, gziptool):
    request = Request(webapp.server.http.base)
    request.add_header('Accept-Encoding', 'gzip')
    opener = build_opener()

    f = opener.open(request)
    s = decompress(f.read())
    assert s == b'Hello World!'


def test2(webapp, gziptool):
    request = Request('%s/static/largefile.txt' % webapp.server.http.base)
    request.add_header('Accept-Encoding', 'gzip')
    opener = build_opener()

    f = opener.open(request)
    s = decompress(f.read())
    assert s == open(path.join(DOCROOT, 'largefile.txt'), 'rb').read()



"""
Serial I/O

This module implements basic Serial (RS232) I/O.
"""

import contextlib
from collections import deque

from circuits.core import Component, Event, handler
from circuits.core.pollers import BasePoller, Poller
from circuits.core.utils import findcmp
from circuits.tools import tryimport

from .events import close, closed, error, opened, read, ready


serial = tryimport('serial')

TIMEOUT = 0.2
BUFSIZE = 4096


class _open(Event):
    """_open Event"""


class Serial(Component):
    channel = 'serial'

    def __init__(
        self, port, baudrate=115200, bufsize=BUFSIZE, timeout=TIMEOUT, encoding='UTF-8', readline=False, channel=channel
    ):
        super().__init__(channel=channel)

        if serial is None:
            raise RuntimeError('No serial support available')

        self._port = port
        self._baudrate = baudrate
        self._bufsize = bufsize
        self._encoding = encoding
        self._readline = readline

        self._serial = None
        self._poller = None
        self._buffer = deque()
        self._closeflag = False

    @handler('ready')
    def _on_ready(self, component):
        self.fire(_open(), self.channel)

    @handler('_open')
    def _on_open(self, port=None, baudrate=None, bufsize=None):
        self._port = port or self._port
        self._baudrate = baudrate or self._baudrate
        self._bufsize = bufsize or self._bufsize

        self._serial = serial.Serial(port=self._port, baudrate=self._baudrate, timeout=0)
        self._fd = self._serial.fileno()  # not portable!

        self._poller.addReader(self, self._fd)

        self.fire(opened(self._port, self._baudrate))

    @handler('registered', 'started', channel='*')
    def _on_registered_or_started(self, component, manager=None):
        if self._poller is None:
            if isinstance(component, BasePoller):
                self._poller = component
                self.fire(ready(self))
            else:
                if component is not self:
                    return
                component = findcmp(self.root, BasePoller)
                if component is not None:
                    self._poller = component
                    self.fire(ready(self))
                else:
                    self._poller = Poller().register(self)
                    self.fire(ready(self))

    @handler('stopped', channel='*')
    def _on_stopped(self, component):
        self.fire(close())

    @handler('prepare_unregister', channel='*')
    def _on_prepare_unregister(self, event, c):
        if event.in_subtree(self):
            self._close()

    def _close(self):
        if self._closeflag:
            return

        self._poller.discard(self._fd)

        self._buffer.clear()
        self._closeflag = False
        self._connected = False

        with contextlib.suppress(OSError):
            self._serial.close()

        self.fire(closed())

    def close(self):
        if not self._buffer:
            self._close()
        elif not self._closeflag:
            self._closeflag = True

    def _read(self):
        try:
            data = self._serial.readline(self._bufsize) if self._readline else self._serial.read(self._bufsize)
            if not isinstance(data, bytes):
                data = data.encode(self._encoding)

            if data:
                self.fire(read(data)).notify = True
        except OSError as exc:
            self.fire(error(exc))
            self._close()

    def _write(self, data):
        try:
            if not isinstance(data, bytes):
                data = data.encode(self._encoding)

            try:
                nbytes = self._serial.write(data)
            except serial.SerialTimeoutException:
                nbytes = 0
            if nbytes < len(data):
                self._buffer.appendleft(data[nbytes:])
        except OSError as e:
            self.fire(error(e))
            self._close()

    def write(self, data):
        if self._poller is not None and not self._poller.isWriting(self._fd):
            self._poller.addWriter(self, self._fd)
        self._buffer.append(data)

    @handler('_disconnect')
    def __on_disconnect(self, sock):
        self._close()

    @handler('_read')
    def __on_read(self, sock):
        self._read()

    @handler('_write')
    def __on_write(self, sock):
        if self._buffer:
            data = self._buffer.popleft()
            self._write(data)

        if not self._buffer:
            if self._closeflag:
                self._close()
            elif self._poller.isWriting(self._fd):
                self._poller.removeWriter(self._fd)



#!/usr/bin/env python
import pytest

from circuits.web import Controller

from .helpers import HTTPError, urlencode, urlopen


class Root(Controller):
    def index(self):
        return 'Hello World!'

    def test_args(self, *args, **kwargs):
        return f'{args!r}\n{kwargs!r}'

    def test_default_args(self, a=None, b=None):
        return f'a={a}\nb={b}'

    def test_redirect(self):
        return self.redirect('/')

    def test_forbidden(self):
        return self.forbidden()

    def test_notfound(self):
        return self.notfound()

    def test_failure(self):
        raise Exception()


def test_root(webapp):
    f = urlopen(webapp.server.http.base)
    s = f.read()
    assert s == b'Hello World!'


def test_404(webapp):
    try:
        urlopen('%s/foo' % webapp.server.http.base)
    except HTTPError as e:
        assert e.code == 404
        assert e.msg == 'Not Found'
    else:
        assert False


def test_args(webapp):
    args = ('1', '2', '3')
    kwargs = {'1': 'one', '2': 'two', '3': 'three'}
    url = '%s/test_args/%s' % (webapp.server.http.base, '/'.join(args))
    data = urlencode(kwargs).encode('utf-8')
    f = urlopen(url, data)
    data = f.read().split(b'\n')
    assert eval(data[0]) == args
    assert eval(data[1]) == kwargs


@pytest.mark.parametrize(
    'data,expected',
    [
        ((['1'], {}), b'a=1\nb=None'),
        ((['1', '2'], {}), b'a=1\nb=2'),
        ((['1'], {'b': '2'}), b'a=1\nb=2'),
    ],
)
def test_default_args(webapp, data, expected):
    args, kwargs = data
    url = '{:s}/test_default_args/{:s}'.format(
        webapp.server.http.base,
        '/'.join(args),
    )
    data = urlencode(kwargs).encode('utf-8')
    f = urlopen(url, data)
    assert f.read() == expected


def test_redirect(webapp):
    f = urlopen('%s/test_redirect' % webapp.server.http.base)
    s = f.read()
    assert s == b'Hello World!'


def test_forbidden(webapp):
    try:
        urlopen('%s/test_forbidden' % webapp.server.http.base)
    except HTTPError as e:
        assert e.code == 403
        assert e.msg == 'Forbidden'
    else:
        assert False


def test_notfound(webapp):
    try:
        urlopen('%s/test_notfound' % webapp.server.http.base)
    except HTTPError as e:
        assert e.code == 404
        assert e.msg == 'Not Found'
    else:
        assert False


def test_failure(webapp):
    try:
        urlopen('%s/test_failure' % webapp.server.http.base)
    except HTTPError as e:
        assert e.code == 500
    else:
        assert False



#!/usr/bin/env python
from optparse import OptionParser

from circuits import Component, Debugger
from circuits.node import Node


__version__ = '0.0.1'

USAGE = '%prog [options]'
VERSION = '%prog v' + __version__


def parse_options():
    parser = OptionParser(usage=USAGE, version=VERSION)

    parser.add_option(
        '-i',
        '--ip',
        action='store',
        type='string',
        default='127.0.0.1:8000',
        dest='bind',
        help='Bind to address:[port]',
    )

    parser.add_option(
        '-d',
        '--debug',
        action='store_true',
        default=False,
        dest='debug',
        help='Enable debug mode',
    )

    opts, args = parser.parse_args()

    return opts, args


class NodeClient(Component):
    def init(self, args, opts):
        if opts.debug:
            Debugger().register(self)

        if ':' in opts.bind:
            address, port = opts.bind.split(':')
            port = int(port)
        else:
            address, port = opts.bind, 8000

        node = Node().register(self)
        node.add('peer_name', address, port)

    def send_all_event(self, infos):
        print('receive: %s' % infos)


def main():
    opts, args = parse_options()

    # Configure and "run" the System.
    NodeClient(args, opts).run()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# curl -i http://localhost:8000/getrange?limit=8
from json import dumps

from circuits.web import Controller, Server


def json(f):
    def wrapper(self, *args, **kwargs):
        return dumps(f(self, *args, **kwargs))

    return wrapper


class Root(Controller):
    @json
    def getrange(self, limit=4):
        return list(range(int(limit)))


app = Server(('0.0.0.0', 8000))
Root().register(app)
app.run()



""".. codeauthor: mnl"""

from signal import SIGINT, SIGTERM
from sys import stderr
from threading import Event
from traceback import format_exception_only

from circuits.core.handlers import reprhandler

from .components import BaseComponent
from .handlers import handler


class FallBackGenerator(BaseComponent):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._continue = Event()

    @handler('generate_events', priority=-100)
    def _on_generate_events(self, event):
        """
        Fall back handler for the :class:`~.events.GenerateEvents` event.

        When the queue is empty a GenerateEvents event is fired, here
        we sleep for as long as possible to avoid using extra cpu cycles.

        A poller would override this with a higher priority handler.
        e.g: ``@handler("generate_events", priority=0)``
        and provide a different way to idle when the queue is empty.
        """
        with event.lock:
            if event.time_left == 0:
                event.stop()
            self._continue.clear()

        if event.time_left > 0:
            # If we get here, there is no component with work to be
            # done and no new event. But some component has requested
            # to be checked again after a certain timeout.
            self._continue.wait(event.time_left)
            # Either time is over or _continue has been set, which
            # implies resume has been called, which means that
            # reduce_time_left(0) has been called. So calling this
            # here is OK in any case.
            event.reduce_time_left(0)
            event.stop()

        while event.time_left < 0:
            # If we get here, there was no work left to do when creating
            # the GenerateEvents event and there is no other handler that
            # is prepared to supply new events within a limited time. The
            # application will continue only if some other Thread fires
            # an event.
            #
            # Python ignores signals when waiting without timeout.
            self._continue.wait(10000)

        event.stop()

    def resume(self):
        """
        Implements the resume method as required from components that
        handle :class:`~.events.GenerateEvents`.
        """
        self._continue.set()


class FallBackExceptionHandler(BaseComponent):
    """
    If there is no handler for error events in the component hierarchy, this
    component's handler is added automatically. It simply prints
    the error information on stderr.
    """

    @handler('exception', channel='*')
    def _on_exception(self, error_type, value, traceback, handler=None, fevent=None):
        s = []

        handler = '' if handler is None else reprhandler(handler)

        msg = f'ERROR {handler} ({fevent!r}) ({error_type!r}): {value!r}\n'

        s.append(msg)
        s.append('Traceback (most recent call last):\n')
        s.extend(traceback)
        s.extend(format_exception_only(error_type, value))
        s.append('\n')
        stderr.write(''.join(s))


class FallBackSignalHandler(BaseComponent):
    """
    If there is no handler for signal events in the component hierarchy, this
    component's handler is added automatically. It simply terminates the
    system if the signal is SIGINT or SIGTERM.
    """

    @handler('signal', channel='*')
    def _on_signal(self, signo, stack):
        if signo in [SIGINT, SIGTERM]:
            raise SystemExit(0)



#!/usr/bin/env python
import os
import signal
from io import StringIO
from subprocess import PIPE, Popen

from circuits import Component, Debugger, Event, handler
from circuits.io import File
from circuits.net.events import write
from circuits.tools import inspect
from circuits.web import Controller, Logger, Server, Sessions, Static
from circuits.web.events import stream


BUFFERING = 1
STREAMING = 2


class kill(Event):
    """kill Event"""


class input(Event):
    """input Event"""


class Command(Component):
    channel = 'cmd'

    def __init__(self, request, response, command, channel=channel):
        super().__init__(channel=channel)

        self._request = request
        self._response = response
        self._command = command

        self._state = BUFFERING
        self._buffer = None

        self._p = Popen(
            command,
            shell=True,
            stdout=PIPE,
            stderr=PIPE,
            close_fds=True,
            preexec_fn=os.setsid,
        )

        self._stdin = None
        if self._p.stdin is not None:
            self._stdin = File(self._p.stdin, channel='%s.stdin' % channel)
            self._stdin.register(self)

        self._stdout = File(self._p.stdout, channel='%s.stdout' % channel)
        self.addHandler(
            handler('eof', channel='%s.stdout' % channel)(self._on_stdout_eof),
        )
        self.addHandler(
            handler('read', channel='%s.stdout' % channel)(
                self._on_stdout_read,
            ),
        )
        self._stdout.register(self)

    @handler('disconnect', channel='web')
    def disconnect(self, sock):
        if sock == self._request.sock:
            self.fire(kill(), self)

    @handler('response', channel='web', priority=-1)
    def response(self, response):
        if response == self._response:
            self._state = STREAMING

    def kill(self):
        os.killpg(self._p.pid, signal.SIGINT)
        self.unregister()

    def input(self, data):
        if self._stdin is not None:
            self.fire(write(data), self._stdin)

    @staticmethod
    def _on_stdout_eof(self):
        if self._buffer is not None:
            self._buffer.flush()
            data = self._buffer.getvalue()
            self.fire(stream(self._response, data), 'web')
        self.fire(stream(self._response, None), 'web')
        self.fire(kill())

    @staticmethod
    def _on_stdout_read(self, data):
        if self._state == BUFFERING:
            if self._buffer is None:
                self._buffer = StringIO()
            self._buffer.write(data)
        elif self._state == STREAMING:
            if self._buffer is not None:
                self._buffer.write(data)
                self._buffer.flush()
                data = self._buffer.getvalue()
                self._buffer = None
                self.fire(stream(self._response, data), 'web')
            else:
                self.fire(stream(self._response, data), 'web')


class Root(Controller):
    def GET(self, *args, **kwargs):
        self.expires(60 * 60 * 24 * 30)
        return self.serve_file(os.path.abspath('static/index.html'))

    def POST(self, input=None):
        if not input:
            return ''

        self.response.headers['Content-Type'] = 'text/plain'

        if input.strip() == 'inspect':
            return inspect(self)

        self.response.stream = True

        sid = self.request.sid
        self += Command(self.request, self.response, input, channel=sid)

        return self.response


app = Server(('0.0.0.0', 8000))
Debugger().register(app)
Static('/js', docroot='static/js').register(app)
Static('/css', docroot='static/css').register(app)
Sessions().register(app)
Logger().register(app)
Root().register(app)
app.run()



import json

from circuits.core import Event


META_EXCLUDE = set(dir(Event()))
META_EXCLUDE.add('node_call_id')
META_EXCLUDE.add('node_sock')
META_EXCLUDE.add('node_without_result')
META_EXCLUDE.add('success_channels')


def load_event(s):
    data = json.loads(s)

    name = data['name']

    args = []
    for arg in data['args']:
        if isinstance(arg, str):
            arg = arg.encode('utf-8')
        args.append(arg)

    kwargs = {}
    for k, v in data['kwargs'].items():
        if isinstance(v, str):
            v = v.encode('utf-8')
        kwargs[str(k)] = v

    e = Event.create(name, *args, **kwargs)

    e.success = bool(data['success'])
    e.failure = bool(data['failure'])
    e.notify = bool(data['notify'])
    e.channels = tuple(data['channels'])

    for k, v in dict(data['meta']).items():
        setattr(e, k, v)

    return e, data['id']


def dump_event(e, id):
    meta = {}
    for name in list(set(dir(e)) - META_EXCLUDE):
        meta[name] = getattr(e, name)

    data = {
        'id': id,
        'name': e.name,
        'args': e.args,
        'kwargs': e.kwargs,
        'success': e.success,
        'failure': e.failure,
        'channels': e.channels,
        'notify': e.notify,
        'meta': meta,
    }

    return json.dumps(data)


def dump_value(v):
    meta = {}
    e = v.event
    if e:
        for name in list(set(dir(e)) - META_EXCLUDE):
            meta[name] = getattr(e, name)

    data = {
        'id': v.node_call_id,
        'errors': v.errors,
        'value': v._value,
        'meta': meta,
    }
    return json.dumps(data)


def load_value(v):
    data = json.loads(v)
    return data['value'], data['id'], data['errors'], data['meta']



#!/usr/bin/env python
import pytest

from circuits import Component, Event, handler


class call(Event):
    """call Event"""

    success = True


class hello(Event):
    """hello Event"""

    success = True


class App(Component):
    @handler('call')
    def _on_call(self):
        x = yield self.call(hello())
        yield x.value

    def hello(self):
        return 'Hello World!'


@pytest.fixture()
def app(request, manager, watcher):
    app = App().register(manager)
    assert watcher.wait('registered')

    def finalizer():
        app.unregister()

    request.addfinalizer(finalizer)

    return app


def test_done_handlers_dont_leak(manager, watcher, app):
    manager.fire(call())
    manager.fire(call())
    assert watcher.wait('call_success')
    assert 'hello_done' not in app._handlers



#!/usr/bin/env python

from circuits import Component, Debugger, Event


class Identify(Event):
    """Identify Event"""

    success = True


class Pound(Component):
    def __init__(self):
        super().__init__()

        Debugger().register(self)
        Bob().register(self)
        Fred().register(self)

    def started(self, *args):
        self.fire(Identify())

    def Identify_success(self, evt, result):
        if not isinstance(result, list):
            result = [result]
        print('In pound:')
        for name in result:
            print(name)


class Dog(Component):
    def Identify(self):
        return self.__class__.__name__


class Bob(Dog):
    """Bob"""


class Fred(Dog):
    """Fred"""


Pound().run()



"""
Errors

This module implements a set of standard HTTP Errors.
"""

import json
import traceback
from html import escape

from circuits import Event

from .constants import DEFAULT_ERROR_MESSAGE, HTTP_STATUS_CODES, POWERED_BY, SERVER_URL, SERVER_VERSION


class httperror(Event):
    """An event for signaling an HTTP error"""

    code = 500
    description = ''

    def __init__(self, request, response, code=None, **kwargs):
        """
        The constructor creates a new instance and modifies the *response*
        argument to reflect the error.
        """
        super().__init__(request, response, code, **kwargs)

        # Override HTTPError subclasses
        self.name = 'httperror'

        self.request = request
        self.response = response

        if code is not None:
            self.code = code

        self.error = kwargs.get('error', None)

        self.description = kwargs.get(
            'description',
            getattr(self.__class__, 'description', ''),
        )

        if self.error is not None:
            stack = self.error[2] if isinstance(self.error[2], (list, tuple)) else traceback.format_tb(self.error[2])
            self.traceback = 'ERROR: (%s) %s\n%s' % (
                self.error[0],
                self.error[1],
                ''.join(stack),
            )
        else:
            self.traceback = ''

        self.response.close = True
        self.response.status = self.code

        powered_by = (
            POWERED_BY
            % {
                'url': escape(SERVER_URL, True),
                'version': escape(SERVER_VERSION),
            }
            if getattr(request.server, 'display_banner', False)
            else ''
        )

        self.data = {
            'code': self.code,
            'name': HTTP_STATUS_CODES.get(self.code, '???'),
            'description': self.description,
            'traceback': self.traceback,
            'powered_by': powered_by,
        }

    def sanitize(self):
        if self.code != 201 and not (299 < self.code < 400) and 'Location' in self.response.headers:
            del self.response.headers['Location']

    def __str__(self):
        self.sanitize()

        if self.code < 200 or self.code in (204, 205, 304):
            return ''

        if 'json' in self.response.headers.get('Content-Type', ''):
            index = ['code', 'name', 'description']
            if self.request.print_debug:
                index.append('traceback')
            return json.dumps({key: self.data[key] for key in index})

        if not self.request.print_debug:
            self.data['traceback'] = ''

        # FIXME: description is a possible XSS attack vector
        return DEFAULT_ERROR_MESSAGE % {
            key: (
                escape(value, True)
                if key not in ('powered_by', 'description') and not isinstance(value, (int, float))
                else value
            )
            for key, value in self.data.items()
        }

    def __repr__(self):
        return '<%s %d %s>' % (
            self.__class__.__name__,
            self.code,
            HTTP_STATUS_CODES.get(
                self.code,
                '???',
            ),
        )


class forbidden(httperror):
    """An event for signaling the HTTP Forbidden error"""

    code = 403


class unauthorized(httperror):
    """An event for signaling the HTTP Unauthorized error"""

    code = 401


class notfound(httperror):
    """An event for signaling the HTTP Not Fouond error"""

    code = 404


class redirect(httperror):
    """An event for signaling the HTTP Redirect response"""

    def __init__(self, request, response, urls, code=None):
        """
        The constructor creates a new instance and modifies the
        *response* argument to reflect a redirect response to the
        given *url*.
        """
        if isinstance(urls, str):
            urls = [urls]

        abs_urls = []
        for url in urls:
            # Note that urljoin will "do the right thing" whether url is:
            #  1. a complete URL with host (e.g. "http://www.example.com/test")
            #  2. a URL relative to root (e.g. "/dummy")
            #  3. a URL relative to the current path
            # Note that any query string in request is discarded.
            url = request.uri.relative(url).unicode()
            abs_urls.append(url)
        self.urls = urls = abs_urls

        # RFC 2616 indicates a 301 response code fits our goal; however,
        # browser support for 301 is quite messy. Do 302/303 instead. See
        # http://ppewww.ph.gla.ac.uk/~flavell/www/post-redirect.html
        if code is None:
            code = 303 if request.protocol >= (1, 1) else 302
        else:
            if code < 300 or code > 399:
                raise ValueError('status code must be between 300 and 399.')

        super().__init__(request, response, code)

        if code in (300, 301, 302, 303, 307, 308):
            response.headers['Content-Type'] = 'text/html'
            # "The ... URI SHOULD be given by the Location field
            # in the response."
            response.headers['Location'] = urls[0]

            # "Unless the request method was HEAD, the entity of the response
            # SHOULD contain a short hypertext note with a hyperlink to the
            # new URI(s)."
            msg = {
                300: "This resource can be found at <a href='%s'>%s</a>.",
                301: ('This resource has permanently moved to <a href="%s">%s</a>.'),
                302: ('This resource resides temporarily at <a href="%s">%s</a>.'),
                303: ('This resource can be found at <a href="%s">%s</a>.'),
                307: ('This resource has moved temporarily to <a href="%s">%s</a>.'),
                308: ('This resource has permanently moved to <a href="%s">%s</a>.'),
            }[code]
            response.body = '<br />\n'.join([msg % (escape(u, True), escape(u)) for u in urls])
            # Previous code may have set C-L, so we have to reset it
            # (allow finalize to set it).
            response.headers.pop('Content-Length', None)
        elif code == 304:
            # Not Modified.
            # "The response MUST include the following header fields:
            # Date, unless its omission is required by section 14.18.1"
            # The "Date" header should have been set in Response.__init__

            # "...the response SHOULD NOT include other entity-headers."
            for key in (
                'Allow',
                'Content-Encoding',
                'Content-Language',
                'Content-Length',
                'Content-Location',
                'Content-MD5',
                'Content-Range',
                'Content-Type',
                'Expires',
                'Last-Modified',
            ):
                if key in response.headers:
                    del response.headers[key]

            # "The 304 response MUST NOT contain a message-body."
            response.body = None
            # Previous code may have set C-L, so we have to reset it.
            response.headers.pop('Content-Length', None)
        elif code == 305:
            # Use Proxy.
            # urls[0] should be the URI of the proxy.
            response.headers['Location'] = urls[0]
            response.body = None
            # Previous code may have set C-L, so we have to reset it.
            response.headers.pop('Content-Length', None)
        else:
            raise ValueError('The %s status code is unknown.' % code)

    def __repr__(self):
        if len(self.channels) > 1:
            channels = repr(self.channels)
        elif len(self.channels) == 1:
            channels = str(self.channels[0])
        else:
            channels = ''
        return '<%s %d[%s.%s] %s>' % (
            self.__class__.__name__,
            self.code,
            channels,
            self.name,
            ' '.join(self.urls),
        )



from circuits import Component, handler
from circuits.core import Value
from circuits.net.events import write

from .utils import dump_event, dump_value, load_event, load_value


DELIMITER = b'~~~'


class Protocol(Component):
    __buffer = b''
    __nid = 0
    __events = {}

    def init(self, sock=None, server=None, **kwargs):
        self.__server = server
        self.__sock = sock
        self.__receive_event_firewall = kwargs.get('receive_event_firewall', None)
        self.__send_event_firewall = kwargs.get('send_event_firewall', None)

    def add_buffer(self, data=''):
        if data:
            self.__buffer += data

        packets = self.__buffer.split(DELIMITER)
        self.__buffer = b''

        for packet in packets:
            try:
                self.__process_packet(packet)
            except ValueError:
                self.__buffer = packet

    @handler(channel='node_result', priority=100)
    def result_handler(self, event, *args, **kwargs):
        if event.name.endswith('_success'):
            source_event = args[0]

            if getattr(args[0], 'node_call_id', False) is not False:
                self.send_result(source_event.node_call_id, source_event.value)

    def send(self, event):
        if self.__send_event_firewall and not self.__send_event_firewall(event, self.__sock):
            yield Value(event, self)

        else:
            id = self.__nid
            self.__nid += 1

            packet = dump_event(event, id).encode('utf-8') + DELIMITER
            self.__send(packet)

            if not getattr(event, 'node_without_result', False):
                self.__events[id] = event
                while not hasattr(self.__events[id], 'remote_finish'):
                    yield

                del self.__events[id]
                yield event.value

    def send_result(self, id, value):
        value.node_call_id = id
        value.node_sock = self.__sock
        packet = dump_value(value).encode('utf-8') + DELIMITER
        self.__send(packet)

    def __send(self, packet):
        if self.__server is not None:
            self.fire(write(self.__sock, packet))
        else:
            self.fire(write(packet))

    def __process_packet(self, packet):
        packet = packet.decode('utf-8')

        if '"value":' in packet:
            self.__process_packet_value(packet)

        else:
            self.__process_packet_call(packet)

    def __process_packet_call(self, packet):
        event, id = load_event(packet)

        if self.__receive_event_firewall and not self.__receive_event_firewall(event, self.__sock):
            self.send_result(id, Value(event, self))
        else:
            event.success = True  # fire %s_success event
            event.success_channels = ('node_result',)
            event.node_call_id = id
            event.node_sock = self.__sock

            # convert byte to str
            event.args = [arg.decode('utf-8') if isinstance(arg, bytes) else arg for arg in event.args]

            for i in event.kwargs:
                v = event.kwargs[i]
                index = i.decode('utf-8') if isinstance(i, bytes) else i
                value = v.decode('utf-8') if isinstance(v, bytes) else v

                del event.kwargs[i]
                event.kwargs[index] = value

            self.fire(event, *event.channels)

    def __process_packet_value(self, packet):
        value, id, error, meta = load_value(packet)

        if id in self.__events:
            # convert byte to str
            value = value.decode('utf-8') if isinstance(value, bytes) else value
            error = error.decode('utf-8') if isinstance(error, bytes) else error

            if not hasattr(self.__events[id], 'value') or not self.__events[id].value:
                self.__events[id].value = Value(self.__events[id], self)

            # save result
            self.__events[id].value.setValue(value)
            self.__events[id].errors = error
            self.__events[id].remote_finish = True

            for k, v in dict(meta).items():
                setattr(self.__events[id], k, v)



"""
Events

This module implements the necessary Events needed.
"""

from circuits import Event


class request(Event):
    """
    request(Event) -> request Event

    args: request, response
    """

    success = True
    failure = True
    complete = True


class response(Event):
    """
    response(Event) -> response Event

    args: request, response
    """

    success = True
    failure = True
    complete = True


class stream(Event):
    """
    stream(Event) -> stream Event

    args: request, response
    """

    success = True
    failure = True
    complete = True


class terminate(Event):
    """terminate Event"""



import base64
import contextlib
import hashlib

from circuits import BaseComponent, handler
from circuits.net.events import connect, disconnect
from circuits.protocols.websocket import WebSocketCodec
from circuits.web.errors import httperror


class WebSocketsDispatcher(BaseComponent):
    """
    This class implements an RFC 6455 compliant WebSockets dispatcher
    that handles the WebSockets handshake and upgrades the connection.

    The dispatcher listens on its channel for :class:`~.web.events.Request`
    events and tries to match them with a given path. Upon a match,
    the request is checked for the proper Opening Handshake information.
    If successful, the dispatcher confirms the establishment of the
    connection to the client. Any subsequent data from the client is
    handled as a WebSocket data frame, decoded and fired as
    a :class:`~.sockets.Read` event on the ``wschannel`` passed to
    the constructor. The data from :class:`~.net.events.write` events on
    that channel is encoded as data frames and forwarded to the client.

    Firing a :class:`~.sockets.Close` event on the ``wschannel`` closes the
    connection in an orderly fashion (i.e. as specified by the
    WebSocket protocol).
    """

    channel = 'web'

    def __init__(self, path=None, wschannel='wsserver', *args, **kwargs):
        """
        :param path: the path to handle. Requests that start with this
            path are considered to be WebSocket Opening Handshakes.

        :param wschannel: the channel on which :class:`~.sockets.read`
            events from the client will be delivered and where
            :class:`~.net.events.write` events to the client will be
            sent to.
        """
        super().__init__(*args, **kwargs)

        self._requests = {}
        self._path = path
        self._wschannel = wschannel
        self._codecs = {}

        @handler('read', channel=wschannel, priority=100)
        def _on_read_handler(self, event, socket, data):
            if socket in self._requests:
                event.request = self._requests[socket]

        self.addHandler(_on_read_handler)

    @handler('request', priority=0.2)
    def _on_request(self, event, request, response):
        if self._path is not None and not request.path.startswith(self._path):
            return None

        self._protocol_version = 13
        headers = request.headers
        sec_key = headers.get('Sec-WebSocket-Key', '').encode('utf-8')
        subprotocols = headers.elements('Sec-WebSocket-Protocol')

        connection_tokens = [s.strip() for s in headers.get('Connection', '').lower().split(',')]

        try:
            if (
                'Host' not in headers
                or headers.get('Upgrade', '').lower() != 'websocket'
                or 'upgrade' not in connection_tokens
                or sec_key is None
                or len(base64.b64decode(sec_key)) != 16
            ):
                return httperror(request, response, code=400)
            if headers.get('Sec-WebSocket-Version', '') != '13':
                response.headers['Sec-WebSocket-Version'] = '13'
                return httperror(request, response, code=400)

            # Generate accept header information
            msg = sec_key + b'258EAFA5-E914-47DA-95CA-C5AB0DC85B11'
            hasher = hashlib.sha1()
            hasher.update(msg)
            accept = base64.b64encode(hasher.digest())

            # Successful completion
            response.status = 101
            response.close = False
            with contextlib.suppress(KeyError):
                del response.headers['Content-Type']
            response.headers['Upgrade'] = 'WebSocket'
            response.headers['Connection'] = 'Upgrade'
            response.headers['Sec-WebSocket-Accept'] = accept.decode('ASCII')
            if subprotocols:
                response.headers['Sec-WebSocket-Protocol'] = self.select_subprotocol(subprotocols)
            codec = WebSocketCodec(request.sock, channel=self._wschannel)
            self._codecs[request.sock] = codec
            codec.register(self)
            return response
        finally:
            event.stop()

    def select_subprotocol(self, subprotocols):
        return subprotocols[0]

    @handler('response_complete')
    def _on_response_complete(self, e, value):
        response = e.args[0]
        request = response.request
        if request.sock in self._codecs:
            self._requests[request.sock] = request
            cevent = connect(
                request.sock,
                *request.sock.getpeername(),
            )
            cevent.request = request
            self.fire(cevent, self._wschannel)

    @handler('disconnect')
    def _on_disconnect(self, sock):
        if sock in self._codecs:
            devent = disconnect(sock)
            devent.request = self._requests.get(sock)
            self.fire(devent, self._wschannel)
            del self._codecs[sock]
            self._requests.pop(sock, None)



#!/usr/bin/env python
import pytest

from circuits import Component, Event, handler


class test(Event):
    """test Event"""


class Base(Component):
    def test(self):
        return 'Hello World!'


class App1(Base):
    @handler('test', priority=-1)
    def test(self):
        return 'Foobar'


class App2(Base):
    @handler('test', override=True)
    def test(self):
        return 'Foobar'


def test_inheritence():
    app = App1()
    app.start()

    x = app.fire(test())
    assert pytest.wait_for(x, 'result')
    v = x.value
    assert v == ['Hello World!', 'Foobar']

    app.stop()


def test_override():
    app = App2()
    app.start()

    x = app.fire(test())
    assert pytest.wait_for(x, 'result')
    v = x.value
    assert v == 'Foobar'

    app.stop()



#!/usr/bin/env python
# Examples:
# curl http://127.0.0.1:8000/ -i -H 'Accept-Encoding: gzip'
# curl http://127.0.0.1:8000/ -i -H 'Accept-Encoding: deflate'
# curl http://127.0.0.1:8000/ -i -H 'Accept-Encoding: identity'

from circuits import Component, Debugger, handler
from circuits.web import Server
from circuits.web.controllers import BaseController, expose
from circuits.web.tools import gzip


class Gzip(Component):
    @handler('response', priority=1.0)
    def compress_response(self, event, response):
        event[0] = gzip(response)


class Root(BaseController):
    @expose('index')
    def index(self):
        return 'Hello World! This is some test string which is compressed using gzip-4 if you want!'


app = Server(('0.0.0.0', 8000))
Root().register(app)
Gzip().register(app)
Debugger().register(app)
app.run()



"""Timer component to facilitate timed events."""

from datetime import datetime
from time import mktime, time

from circuits.core.handlers import handler

from .components import BaseComponent


class Timer(BaseComponent):
    """
    Timer Component

    A timer is a component that fires an event once after a certain
    delay or periodically at a regular interval.
    """

    def __init__(self, interval, event, *channels, **kwargs):
        """
        :param interval: the delay or interval to wait for until
                         the event is fired. If interval is specified as
                         datetime, the interval is recalculated as the
                         time span from now to the given datetime.
        :type interval:  ``datetime`` or number of seconds as a ``float``

        :param event:    the event to fire.
        :type event:     :class:`~.events.Event`

        :param persist:  An optional keyword argument which if ``True``
                         will cause the event to be fired repeatedly
                         once per configured interval until the timer
                         is unregistered.  If ``False``, the event fires
                         exactly once after the specified interval, and
                         the timer is unregistered. **Default:** ``False``
        :type persist:   ``bool``
        """
        super().__init__()

        self.expiry = None
        self.interval = None
        self.event = event
        self.channels = channels
        self.persist = kwargs.get('persist', False)

        self.reset(interval)

    @handler('generate_events')
    def _on_generate_events(self, event):
        if self.expiry is None:
            return

        now = time()

        if now >= self.expiry:
            if self.unregister_pending:
                return
            self.fire(self.event, *self.channels)

            if self.persist:
                self.reset()
            else:
                self.unregister()
            event.reduce_time_left(0)
        else:
            event.reduce_time_left(self.expiry - now)

    def reset(self, interval=None):
        """
        Reset the timer, i.e. clear the amount of time already waited
        for.
        """
        if interval is not None and isinstance(interval, datetime):
            self.interval = mktime(interval.timetuple()) - time()
        elif interval is not None:
            self.interval = interval

        self.expiry = time() + self.interval

    @property
    def expiry(self):
        return getattr(self, '_expiry', None)

    @expiry.setter
    def expiry(self, seconds):
        self._expiry = seconds



#!/usr/bin/env python
from io import BytesIO
from os import path

import pytest

from circuits.web import Controller

from .helpers import Request, urlopen
from .multipartform import MultiPartForm


@pytest.fixture()
def sample_file(request):
    return open(
        path.join(
            path.dirname(__file__),
            'static',
            'unicode.txt',
        ),
        'rb',
    )


class Root(Controller):
    def index(self, file, description=''):
        yield 'Filename: %s\n' % file.filename
        yield 'Description: %s\n' % description
        yield 'Content:\n'
        yield file.value

    def upload(self, file, description=''):
        return file.value


def test(webapp):
    form = MultiPartForm()
    form['description'] = 'Hello World!'

    fd = BytesIO(b'Hello World!')
    form.add_file('file', 'helloworld.txt', fd, 'text/plain; charset=utf-8')

    # Build the request
    url = webapp.server.http.base
    data = form.bytes()
    headers = {
        'Content-Type': form.get_content_type(),
        'Content-Length': len(data),
    }

    request = Request(url, data, headers)

    f = urlopen(request)
    s = f.read()
    lines = s.split(b'\n')

    assert lines[0] == b'Filename: helloworld.txt'
    assert lines[1] == b'Description: Hello World!'
    assert lines[2] == b'Content:'
    assert lines[3] == b'Hello World!'


def test_unicode(webapp, sample_file):
    form = MultiPartForm()
    form['description'] = sample_file.name
    form.add_file(
        'file',
        'helloworld.txt',
        sample_file,
        'text/plain; charset=utf-8',
    )

    # Build the request
    url = f'{webapp.server.http.base:s}/upload'
    data = form.bytes()
    headers = {
        'Content-Type': form.get_content_type(),
        'Content-Length': len(data),
    }

    request = Request(url, data, headers)

    f = urlopen(request)
    s = f.read()
    sample_file.seek(0)
    expected_output = sample_file.read()  # use the byte stream
    assert s == expected_output



#!/usr/bin/env python
from datetime import datetime
from email.utils import parsedate
from time import mktime

from circuits.web import Controller

from .helpers import urlopen


class Root(Controller):
    def index(self):
        self.expires(60)
        return 'Hello World!'

    def nocache(self):
        self.expires(0)
        return 'Hello World!'


def test(webapp):
    f = urlopen(webapp.server.http.base)
    s = f.read()
    assert s == b'Hello World!'
    expires = f.headers['Expires']
    diff = mktime(parsedate(expires)) - mktime(datetime.utcnow().timetuple())
    assert 60 - (60 * 0.1) < diff < 60 + (60 * 0.1)  # diff is about 60 +- 10%


def test_nocache(webapp):
    f = urlopen('%s/nocache' % webapp.server.http.base)
    s = f.read()
    assert s == b'Hello World!'

    expires = f.headers['Expires']
    pragma = f.headers['Pragma']
    cacheControl = f.headers['Cache-Control']

    now = datetime.utcnow()
    lastyear = now.replace(year=now.year - 1)

    diff = mktime(parsedate(expires)) - mktime(lastyear.utctimetuple())
    assert diff < 1.0

    assert pragma == 'no-cache'
    assert cacheControl == 'no-cache, must-revalidate'



#!/usr/bin/env python
from circuits.web import Controller
from circuits.web.wsgi import Application


class Root(Controller):
    def index(self):
        return 'Hello World!'


application = Application()
Root().register(application)



#!/usr/bin/env python
from .helpers import urlopen


def application(environ, start_response):
    status = '200 OK'
    response_headers = [('Content-type', 'text/plain')]
    start_response(status, response_headers)
    return 'Hello World!'


def test(webapp):
    f = urlopen(webapp.server.http.base)
    s = f.read()
    assert s == b'Hello World!'
    assert f.headers.get('Transfer-Encoding') != 'chunked'



#!/usr/bin/env python
from circuits import Component, Event, handler


class foo(Event):
    """foo Event"""


class test(Event):
    """test Event"""


class A(Component):
    channel = 'a'

    def test(self):
        return 'Hello World!'

    @handler(priority=1.0)
    def _on_event(self, event, *args, **kwargs):
        return 'Foo'


class B(Component):
    @handler(priority=10.0, channel='*')
    def _on_channel(self, event, *args, **kwargs):
        return 'Bar'


def test_main():
    app = A() + B()
    while len(app):
        app.flush()

    x = app.fire(test(), 'a')
    while len(app):
        app.flush()

    assert x.value[0] == 'Bar'
    assert x.value[1] == 'Foo'
    assert x.value[2] == 'Hello World!'


def test_event():
    app = A() + B()
    while len(app):
        app.flush()

    e = test()
    x = app.fire(e)
    while len(app):
        app.flush()

    assert x.value[0] == 'Bar'
    assert x.value[1] == 'Foo'
    assert x.value[2] == 'Hello World!'


def test_channel():
    app = A() + B()
    while len(app):
        app.flush()

    e = foo()
    x = app.fire(e, 'b')
    while len(app):
        app.flush()

    assert x.value == 'Bar'



#!/usr/bin/env python
import os
import sys
from errno import ESRCH
from os import kill, remove
from signal import SIGTERM
from subprocess import Popen
from time import sleep

import pytest

from . import signalapp


def is_running(pid):
    try:
        kill(pid, 0)
    except OSError as error:
        if error.errno == ESRCH:
            return False
    return True


def wait(pid, timeout=3):
    count = timeout
    while is_running(pid) and count:
        sleep(1)


def test(tmpdir):
    if os.name != 'posix':
        pytest.skip('Cannot run test on a non-POSIX platform.')

    tmpdir.ensure('.pid')
    tmpdir.ensure('.signal')
    pidfile = str(tmpdir.join('.pid'))
    signalfile = str(tmpdir.join('.signal'))

    args = [sys.executable, signalapp.__file__, pidfile, signalfile]
    cmd = ' '.join(args)
    p = Popen(cmd, shell=True, env={'PYTHONPATH': ':'.join(sys.path)})
    status = p.wait()

    assert status == 0

    sleep(1)

    assert os.path.exists(pidfile)
    assert os.path.isfile(pidfile)

    f = open(pidfile)
    pid = int(f.read().strip())
    f.close()

    kill(pid, SIGTERM)
    wait(pid)

    with open(signalfile) as fd:
        signal = fd.read().strip()

    assert int(signal) == int(SIGTERM)

    remove(pidfile)
    remove(signalfile)



#!/usr/bin/env python
from hashlib import md5

from circuits import Component, handler
from circuits.web import Controller, Server
from circuits.web.tools import basic_auth, check_auth


class Auth(Component):
    realm = 'Test'
    users = {'admin': md5(b'admin').hexdigest()}

    @handler('request', priority=1.0)
    def on_request(self, event, request, response):
        """
        Filter Requests applying Basic Authentication

        Filter any incoming requests at a higher priority than the
        default dispatcher and apply Basic Authentication returning
        a 403 Forbidden response if Authentication failed.
        """
        if not check_auth(request, response, self.realm, self.users):
            event.stop()
            return basic_auth(request, response, self.realm, self.users)
        return None


class Root(Controller):
    def index(self):
        return 'Hello World!'


app = Server(('0.0.0.0', 8000))
Auth().register(app)
Root().register(app)
app.run()



import itertools
from email.generator import _make_boundary
from mimetypes import guess_type


class MultiPartForm(dict):
    def __init__(self):
        self.files = []
        self.boundary = _make_boundary()

    def get_content_type(self):
        return 'multipart/form-data; boundary="%s"' % self.boundary

    def add_file(self, fieldname, filename, fd, mimetype=None):
        body = fd.read()
        if mimetype is None:
            mimetype = guess_type(filename)[0] or 'application/octet-stream'
        self.files.append((fieldname, filename, mimetype, body))

    def bytes(self):
        parts = []
        part_boundary = bytearray('--%s' % self.boundary, 'ascii')

        # Add the form fields
        parts.extend(
            [
                part_boundary,
                bytearray(
                    'Content-Disposition: form-data; name="%s"' % k,
                    'ascii',
                ),
                b'',
                v if isinstance(v, bytes) else bytearray(v, 'ascii'),
            ]
            for k, v in list(self.items())
        )

        # Add the files to upload
        parts.extend(
            [
                part_boundary,
                bytearray(
                    'Content-Disposition: form-data; name="%s"; filename="%s"' % (fieldname, filename),
                    'ascii',
                ),
                bytearray('Content-Type: %s' % content_type, 'ascii'),
                bytearray(),
                body if isinstance(body, bytes) else bytearray(body, 'ascii'),
            ]
            for fieldname, filename, content_type, body in self.files
        )

        # Flatten the list and add closing boundary marker,
        # then return CR+LF separated data
        flattened = list(itertools.chain(*parts))
        flattened.append(bytearray('--%s--' % self.boundary, 'ascii'))
        res = bytearray()
        for item in flattened:
            res += item
            res += bytearray('\r\n', 'ascii')
        return res



#!/usr/bin/env python

from circuits import BaseComponent, Debugger, handler


class MyComponent(BaseComponent):
    def __init__(self):
        super().__init__()

        Debugger().register(self)

    @handler('started', channel='*')
    def system_started(self, component):
        print('Start event detected')


MyComponent().run()



#!/usr/bin/env python
from circuits import Component, Event


class foo(Event):
    """foo Event"""


class done(Event):
    """done Event"""


class App(Component):
    def init(self):
        self.results = []

    def foo(self, value):
        self.results.append(value)

    def done(self):
        self.stop()


def test1():
    app = App()

    # Normal Order
    [app.fire(foo(1)), app.fire(foo(2))]  # noqa: B018
    app.fire(done())

    app.run()

    assert app.results == [1, 2]


def test2():
    app = App()

    # Priority Order
    [app.fire(foo(1), priority=2), app.fire(foo(2), priority=0)]  # noqa: B018
    app.fire(done())

    app.run()

    assert app.results == [2, 1]



#!/usr/bin/env python
from circuits import Manager
from circuits.core.pollers import BasePoller, Poller
from circuits.core.utils import findtype
from circuits.net.sockets import TCPClient, TCPServer


def test():
    m = Manager()

    poller = Poller().register(m)

    TCPServer(0).register(m)
    TCPClient().register(m)

    m.start()

    try:
        pollers = findtype(m, BasePoller, all=True)
        assert len(pollers) == 1
        assert pollers[0] is poller
    finally:
        m.stop()



#!/usr/bin/env python
from circuits.web import Controller
from circuits.web.client import Client, request


class Root(Controller):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self += Leaf()

    def index(self):
        return 'Hello World!'

    def name(self):
        return 'Earth'


class Leaf(Controller):
    channel = '/world/country/region'

    def index(self):
        return 'Hello cities!'

    def city(self):
        return 'Hello City!'


def make_request(webapp, path):
    client = Client()
    client.start()

    client.fire(request('GET', path))
    while client.response is None:
        pass

    client.stop()

    response = client.response
    s = response.read()
    return response.status, s


def test_root(webapp):
    status, content = make_request(webapp, webapp.server.http.base)

    assert status == 200
    assert content == b'Hello World!'


def test_root_name(webapp):
    status, content = make_request(webapp, '%s/name' % webapp.server.http.base)

    assert status == 200
    assert content == b'Earth'


def test_leaf(webapp):
    status, content = make_request(webapp, '%s/world/country/region' % webapp.server.http.base)

    assert status == 200
    assert content == b'Hello cities!'


def test_city(webapp):
    status, content = make_request(webapp, '%s/world/country/region/city' % webapp.server.http.base)

    assert status == 200
    assert content == b'Hello City!'



"""Workers Tests"""

import pytest

from circuits import Worker, task


task.complete = True


@pytest.fixture()
def worker(request, manager, watcher):
    worker = Worker().register(manager)
    assert watcher.wait('registered')

    def finalizer():
        worker.unregister()
        assert watcher.wait('unregistered')

    request.addfinalizer(finalizer)

    return worker


def f():
    x = 0
    i = 0
    while i < 1000000:
        x += 1
        i += 1
    return x


def add(a, b):
    return a + b


def test(manager, watcher, worker):
    x = manager.fire(task(f))
    assert watcher.wait('task_complete')

    assert x.result
    assert x.value == 1000000


def test_args(manager, watcher, worker):
    x = manager.fire(task(add, 1, 2))
    assert watcher.wait('task_complete')

    assert x.result
    assert x.value == 3



import re
from cgi import parse_header

from .headers import HeaderElement
from .parsers import MultipartParser, QueryStringParser


def process_multipart(request, params):
    headers = request.headers

    ctype = headers.elements('Content-Type')
    ctype = ctype[0] if ctype else HeaderElement.from_str('application/x-www-form-urlencoded')

    ib = ''
    if 'boundary' in ctype.params:
        # http://tools.ietf.org/html/rfc2046#section-5.1.1
        # "The grammar for parameters on the Content-type field is such that it
        # is often necessary to enclose the boundary parameter values in quotes
        # on the Content-type line"
        ib = ctype.params['boundary'].strip('"')

    if not re.match('^[ -~]{0,200}[!-~]$', ib):
        raise ValueError(f'Invalid boundary in multipart form: {ib!r}')

    parser = MultipartParser(request.body, ib)
    for part in parser:
        if part.filename or not part.is_buffered():
            params[part.name] = part
        else:
            params[part.name] = part.value


def process_urlencoded(request, params, encoding='utf-8'):
    params.update(QueryStringParser(request.qs).result)
    body = request.body.getvalue().decode(encoding)
    result = QueryStringParser(body).result
    for key, value in result.items():
        params[_decode_value(key, encoding)] = _decode_value(value, encoding)


def _decode_value(value, encoding):
    if isinstance(value, bytes):
        value = value.decode(encoding)
    elif isinstance(value, list):
        value = [_decode_value(val, encoding) for val in value]
    elif isinstance(value, dict):
        value = {key.decode(encoding): _decode_value(val, encoding) for key, val in value.iteritems()}
    return value


def process(request, params):
    ctype = request.headers.get('Content-Type')
    if not ctype:
        return

    mtype, mencoding = ctype.split('/', 1) if '/' in ctype else (ctype, None)
    mencoding, extra = parse_header(mencoding)

    charset = extra.get('charset', 'utf-8')

    if mtype == 'multipart':
        process_multipart(request, params)
    elif mtype == 'application' and mencoding == 'x-www-form-urlencoded':
        process_urlencoded(request, params, encoding=charset)



#!/usr/bin/env python

import pytest

from circuits import Component
from circuits.net.sockets import BUFSIZE, close, write
from circuits.web.controllers import Controller
from circuits.web.websockets import WebSocketClient, WebSocketsDispatcher

from .helpers import urlopen


class Echo(Component):
    channel = 'wsserver'

    def init(self):
        self.clients = []

    def connect(self, sock, host, port):
        self.clients.append(sock)
        print('WebSocket Client Connected:', host, port)
        self.fire(write(sock, f'Welcome {host:s}:{port:d}'))

    def disconnect(self, sock):
        self.clients.remove(sock)

    def read(self, sock, data):
        self.fire(write(sock, 'Received: ' + data))


class Root(Controller):
    def index(self):
        return 'Hello World!'


class Client(Component):
    channel = 'ws'

    def init(self, *args, **kwargs):
        self.response = None

    def read(self, data):
        self.response = data


@pytest.mark.parametrize('chunksize', [BUFSIZE, BUFSIZE + 1, BUFSIZE * 2])
def test(manager, watcher, webapp, chunksize):
    echo = Echo().register(webapp)
    assert watcher.wait('registered', channel='wsserver')

    f = urlopen(webapp.server.http.base)
    s = f.read()
    assert s == b'Hello World!'

    watcher.clear()

    WebSocketsDispatcher('/websocket').register(webapp)
    assert watcher.wait('registered', channel='web')

    uri = f'ws://{webapp.server.host:s}:{webapp.server.port:d}/websocket'

    WebSocketClient(uri).register(manager)
    client = Client().register(manager)
    assert watcher.wait('registered', channel='wsclient')
    assert watcher.wait('connected', channel='wsclient')

    assert watcher.wait('connect', channel='wsserver')
    assert len(echo.clients) == 1

    assert watcher.wait('read', channel='ws')
    assert client.response.startswith('Welcome')
    watcher.clear()

    client.fire(write('Hello!'), 'ws')
    assert watcher.wait('read', channel='ws')
    assert client.response == 'Received: Hello!'

    watcher.clear()

    client.fire(write('World!'), 'ws')
    assert watcher.wait('read', channel='ws')
    assert client.response == 'Received: World!'

    watcher.clear()

    data = 'A' * (chunksize + 1)
    client.fire(write(data), 'ws')
    assert watcher.wait('read', channel='ws')
    assert client.response == f'Received: {data}'

    f = urlopen(webapp.server.http.base)
    s = f.read()
    assert s == b'Hello World!'

    assert len(echo.clients) == 1

    client.fire(close(), 'ws')
    assert watcher.wait('disconnect', channel='wsserver')
    assert len(echo.clients) == 0

    client.unregister()
    assert watcher.wait('unregistered')



#!/usr/bin/env python

from circuits import Component, Event


class woof(Event):
    """woof Event"""


class Pound(Component):
    def __init__(self):
        super().__init__()

        self.bob = Bob().register(self)
        self.fred = Fred().register(self)

    def started(self, *args):
        self.fire(woof())


class Dog(Component):
    def woof(self):
        print("Woof! I'm %s!" % self.name)


class Bob(Dog):
    """Bob"""


class Fred(Dog):
    """Fred"""


Pound().run()



"""Debugger Tests"""

import sys
from io import StringIO

import pytest

from circuits import Debugger
from circuits.core import Component, Event


class test(Event):
    """test Event"""


class App(Component):
    def test(self, raiseException=False):
        if raiseException:
            raise Exception()


class Logger:
    error_msg = None
    debug_msg = None

    def error(self, msg):
        self.error_msg = msg

    def debug(self, msg):
        self.debug_msg = msg


def test_main():
    app = App()
    stderr = StringIO()
    debugger = Debugger(file=stderr)
    debugger.register(app)
    while len(app):
        app.flush()
    stderr.seek(0)
    stderr.truncate()

    assert debugger._events

    e = Event()
    app.fire(e)
    app.flush()
    stderr.seek(0)
    s = stderr.read().strip()
    assert s == str(e)
    stderr.seek(0)
    stderr.truncate()

    debugger._events = False
    assert not debugger._events

    e = Event()
    app.fire(e)

    stderr.seek(0)
    s = stderr.read().strip()
    assert s == ''
    stderr.seek(0)
    stderr.truncate()


def test_file(tmpdir):
    logfile = str(tmpdir.ensure('debug.log'))
    stderr = open(logfile, 'w+')

    app = App()
    debugger = Debugger(file=stderr)
    debugger.register(app)
    while len(app):
        app.flush()
    stderr.seek(0)
    stderr.truncate()

    assert debugger._events

    e = Event()
    app.fire(e)
    app.flush()

    stderr.seek(0)
    s = stderr.read().strip()
    assert s == str(e)
    stderr.seek(0)
    stderr.truncate()

    debugger._events = False
    assert not debugger._events

    e = Event()
    app.fire(e)

    stderr.seek(0)
    s = stderr.read().strip()
    assert s == ''
    stderr.seek(0)
    stderr.truncate()


def test_filename(tmpdir):
    if '__pypy__' in sys.modules:
        pytest.skip('Broken on pypy')

    logfile = str(tmpdir.ensure('debug.log'))
    stderr = open(logfile, 'r+')

    app = App()
    debugger = Debugger(file=logfile)
    debugger.register(app)
    while len(app):
        app.flush()
    stderr.seek(0)
    stderr.truncate()

    assert debugger._events

    e = Event()
    app.fire(e)
    app.flush()

    stderr.seek(0)
    s = stderr.read().strip()
    assert s == str(e)
    stderr.seek(0)
    stderr.truncate()

    debugger._events = False
    assert not debugger._events

    e = Event()
    app.fire(e)

    stderr.seek(0)
    s = stderr.read().strip()
    assert s == ''
    stderr.seek(0)
    stderr.truncate()


def test_exceptions():
    app = App()
    stderr = StringIO()
    debugger = Debugger(file=stderr)
    debugger.register(app)
    while len(app):
        app.flush()
    stderr.seek(0)
    stderr.truncate()

    assert debugger._events
    assert debugger._errors

    e = test(raiseException=True)
    app.fire(e)
    app.flush()

    stderr.seek(0)
    s = stderr.read().strip()
    assert s == str(e)
    stderr.seek(0)
    stderr.truncate()

    app.flush()
    stderr.seek(0)
    s = stderr.read().strip()
    assert s.startswith('<exception[*]')
    stderr.seek(0)
    stderr.truncate()

    debugger._events = False
    debugger._errors = False

    assert not debugger._events
    assert not debugger._errors

    e = test(raiseException=True)
    app.fire(e)
    app.flush()

    stderr.seek(0)
    s = stderr.read().strip()
    assert s == ''
    stderr.seek(0)
    stderr.truncate()

    app.flush()
    stderr.seek(0)
    s = stderr.read().strip()
    assert s == ''


def test_IgnoreEvents():
    app = App()
    stderr = StringIO()
    debugger = Debugger(file=stderr)
    debugger.register(app)
    while len(app):
        app.flush()
    stderr.seek(0)
    stderr.truncate()

    assert debugger._events

    debugger.IgnoreEvents.extend(['test'])

    e = Event()
    app.fire(e)
    app.flush()

    stderr.seek(0)
    s = stderr.read().strip()
    assert s == str(e)
    stderr.seek(0)
    stderr.truncate()

    e = test()
    app.fire(e)
    app.flush()

    stderr.seek(0)
    s = stderr.read().strip()
    assert s == ''
    stderr.seek(0)
    stderr.truncate()


def test_IgnoreChannels():
    app = App()
    stderr = StringIO()
    debugger = Debugger(file=stderr)
    debugger.register(app)
    while len(app):
        app.flush()
    stderr.seek(0)
    stderr.truncate()

    assert debugger._events
    debugger.IgnoreChannels.extend([('*', 'test')])

    e = Event()
    app.fire(e)
    app.flush()

    stderr.seek(0)
    s = stderr.read().strip()
    assert s == str(e)
    stderr.seek(0)
    stderr.truncate()

    e = test()
    app.fire(e)
    app.flush()

    stderr.seek(0)
    s = stderr.read().strip()
    assert s == ''
    stderr.seek(0)
    stderr.truncate()


def test_Logger_debug():
    app = App()
    logger = Logger()
    debugger = Debugger(logger=logger)
    debugger.register(app)
    while len(app):
        app.flush()

    e = Event()
    app.fire(e)
    app.flush()

    assert logger.debug_msg == repr(e)


def test_Logger_error():
    app = App()
    logger = Logger()
    debugger = Debugger(logger=logger)
    debugger.register(app)
    while len(app):
        app.flush()

    e = test(raiseException=True)
    app.fire(e)
    while len(app):
        app.flush()

    assert logger.error_msg.startswith('ERROR <handler[*][test] (App.test)> (')



#!/usr/bin/env python
from circuits.web import Controller

from .helpers import urlopen


class Root(Controller):
    def index(self, *args, **kwargs):
        return 'ERROR'


def application(environ, start_response):
    status = '200 OK'
    start_response(status, [])
    return ['']


def test(webapp):
    f = urlopen(webapp.server.http.base)
    s = f.read()
    assert s == b''
    assert f.headers.get('Transfer-Encoding') != 'chunked'



from circuits.net.sockets import BUFSIZE
from circuits.web.servers import BaseServer


class MockClass(BaseServer):
    pass


def test_dynamic_bufsize_in_baseserver():
    bufsize = 10000
    try:
        # Assert when we set BUFSIZE ourselves
        mock = MockClass(bind='0.0.0.0:1234', bufsize=bufsize)
        # that it will be set to our given value
        assert bufsize == mock.server._bufsize
    finally:
        mock.server.unregister()
        mock.stop()


def test_constant_bufsize_in_baseserver():
    try:
        # Assert when we dont set BUFSIZE ourself
        mock = MockClass(bind='0.0.0.0:1235')
        # that it will be set to the constant default value
        assert mock.server._bufsize == BUFSIZE
    finally:
        mock.server.unregister()
        mock.stop()



#!/usr/bin/env python
"""
Chat Server Example

This example demonstrates how to create a very simple telnet-style chat
server that supports many connecting clients.
"""

from optparse import OptionParser

from circuits import Component, Debugger
from circuits.net.events import write
from circuits.net.sockets import TCPServer


__version__ = '0.0.1'

USAGE = '%prog [options]'
VERSION = '%prog v' + __version__


def parse_options():
    parser = OptionParser(usage=USAGE, version=VERSION)

    parser.add_option(
        '-b',
        '--bind',
        action='store',
        type='string',
        default='0.0.0.0:8000',
        dest='bind',
        help='Bind to address:[port]',
    )

    parser.add_option(
        '-d',
        '--debug',
        action='store_true',
        default=False,
        dest='debug',
        help='Enable debug mode',
    )

    opts, args = parser.parse_args()

    return opts, args


class ChatServer(Component):
    def init(self, args, opts):
        """
        Initialize our ``ChatServer`` Component.

        This uses the convenience ``init`` method which is called after the
        component is properly constructed and initialized and passed the
        same args and kwargs that were passed during construction.
        """
        self.args = args
        self.opts = opts

        self.clients = {}

        if opts.debug:
            Debugger().register(self)

        if ':' in opts.bind:
            address, port = opts.bind.split(':')
            port = int(port)
        else:
            address, port = opts.bind, 8000

        bind = (address, port)

        TCPServer(bind).register(self)

    def broadcast(self, data, exclude=None):
        exclude = exclude or []
        targets = (sock for sock in self.clients if sock not in exclude)
        for target in targets:
            self.fire(write(target, data))

    def connect(self, sock, host, port):
        """Connect Event -- Triggered for new connecting clients"""
        self.clients[sock] = {
            'host': sock,
            'port': port,
            'state': {
                'nickname': None,
                'registered': False,
            },
        }

        self.fire(write(sock, b'Welcome to the circuits Chat Server!\n'))
        self.fire(write(sock, b'Please enter a desired nickname: '))

    def disconnect(self, sock):
        """Disconnect Event -- Triggered for disconnecting clients"""
        if sock not in self.clients:
            return

        nickname = self.clients[sock]['state']['nickname']

        self.broadcast(
            f'!!! {nickname:s} has left !!!\n'.encode(),
            exclude=[sock],
        )

        del self.clients[sock]

    def read(self, sock, data):
        """Read Event -- Triggered for when client connections have data"""
        data = data.strip().decode('utf-8')

        if not self.clients[sock]['state']['registered']:
            nickname = data
            self.clients[sock]['state']['registered'] = True
            self.clients[sock]['state']['nickname'] = nickname

            self.broadcast(
                f'!!! {nickname:s} has joined !!!\n'.encode(),
                exclude=[sock],
            )
        else:
            nickname = self.clients[sock]['state']['nickname']

            self.broadcast(
                f'<{nickname:s}> {data:s}\n'.encode(),
                exclude=[sock],
            )


def main():
    opts, args = parse_options()

    # Configure and "run" the System.
    ChatServer(args, opts).run()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
Test Interface Query

Test the capabilities of querying a Component class or instance for it's
interface. That is it's event handlers it responds to.
"""

from circuits import Component


class Base(Component):
    def foo(self):
        pass


class SuperBase(Base):
    def bar(self):
        pass


def test_handles_base_class():
    assert Base.handles('foo')


def test_handles_super_base_class():
    assert SuperBase.handles('foo', 'bar')


def test_handles_base_instance():
    base = Base()
    assert base.handles('foo')


def test_handles_super_base_instance():
    superbase = SuperBase()
    assert superbase.handles('foo', 'bar')



#!/usr/bin/env python
from collections import defaultdict

from circuits import Component, Event
from circuits.protocols import Line


class read(Event):
    """read Event"""


class App(Component):
    lines = []

    def line(self, line):
        self.lines.append(line)


class AppServer(Component):
    channel = 'server'

    lines = []

    def line(self, sock, line):
        self.lines.append((sock, line))


def test():
    app = App()
    Line().register(app)

    while len(app):
        app.flush()

    app.fire(read(b'1\n2\r\n3\n4'))

    while len(app):
        app.flush()

    assert app.lines[0] == b'1'
    assert app.lines[1] == b'2'
    assert app.lines[2] == b'3'


def test_server():
    app = AppServer()
    buffers = defaultdict(bytes)
    Line(
        getBuffer=buffers.__getitem__,
        updateBuffer=buffers.__setitem__,
    ).register(app)

    while len(app):
        app.flush()

    app.fire(read(1, b'1\n2\r\n3\n4'))
    app.fire(read(2, b'1\n2\r\n3\n4'))

    while len(app):
        app.flush()

    assert app.lines[0] == (1, b'1')
    assert app.lines[1] == (1, b'2')
    assert app.lines[2] == (1, b'3')

    assert app.lines[3] == (2, b'1')
    assert app.lines[4] == (2, b'2')
    assert app.lines[5] == (2, b'3')



"""Utilities"""

from contextlib import contextmanager
from functools import wraps
from imp import find_module

from fabric.api import abort, hide, local, puts, quiet, settings, warn


def tobool(s):
    if isinstance(s, bool):
        return s
    return s.lower() in ['yes', 'y']


def toint(s):
    if isinstance(s, int):
        return s
    return int(s)


@contextmanager
def msg(s):
    """
    Print message given as ``s`` in a context manager

    Prints "{s} ... OK"
    """
    puts(f'{s:s} ... ', end='', flush=True)
    with settings(hide('everything')):
        yield
    puts('OK', show_prefix=False, flush=True)


def pip(*args, **kwargs):
    requirements = kwargs.get('requirements', None)
    if requirements is not None:
        local('pip install -U -r {:s}'.format(kwargs['requirements']))
    else:
        args = [arg for arg in args if not has_module(arg)]
        if args:
            local('pip install {:s}'.format(' '.join(args)))


def has_module(name):
    try:
        return find_module(name)
    except ImportError:
        return False


def has_binary(name):
    with quiet():
        return local(f'which {name:s}').succeeded


def requires(*names, **kwargs):
    """
    Decorator/Wrapper that aborts if not all requirements are met.

    Aborts if not all requirements are met given a test function
    (defaulting to :func:`~has_binary`).

    :param kwargs: Optional kwargs. e.g: ``test=has_module``
    :type kwargs: dict

    :returns: None or aborts
    :rtype: None
    """
    test = kwargs.get('test', has_binary)

    def decorator(f):
        @wraps(f)
        def wrapper(*args, **kwds):
            if all(test(name) for name in names):
                return f(*args, **kwds)
            for name in names:
                if not test(name):
                    warn(f'{name:s} not found')
            abort(f'requires({names!r:s}) failed')
            return None

        return wrapper

    return decorator



#!/usr/bin/env python
from os.path import dirname

import pytest

from circuits import Event, Loader, Manager


class test(Event):
    """test Event"""


def test_main():
    m = Manager()
    loader = Loader(paths=[dirname(__file__)]).register(m)

    m.start()

    loader.load('app')

    x = m.fire(test())

    assert pytest.wait_for(x, 'result')

    s = x.value
    assert s == 'Hello World!'

    m.stop()



#!/usr/bin/env python
from pytest import PLATFORM, fixture, skip

from circuits import Component, Event
from circuits.net.events import close
from circuits.net.sockets import UDPServer
from circuits.node import Node, remote


if PLATFORM == 'win32':
    skip('Broken on Windows')


class App(Component):
    _ready = False
    value = False
    disconnected = False

    def foo(self):
        return 'Hello World!'

    def ready(self, *args):
        self._ready = True

    def disconnect(self, component):
        self.disconnected = True

    def remote_value_changed(self, value):
        self.value = True


@fixture()
def bind(request, manager, watcher):
    server = UDPServer(0).register(manager)
    assert watcher.wait('ready')

    host, port = server.host, server.port

    server.fire(close())
    assert watcher.wait('closed')

    server.unregister()
    assert watcher.wait('unregistered')

    return host, port


@fixture()
def app(request, manager, watcher, bind):
    app = App().register(manager)
    node = Node().register(app)
    watcher.wait('ready')

    child = App() + Node(port=bind[1], server_ip=bind[0])
    child.start(process=True)

    node.add('child', *bind)
    watcher.wait('connected')

    def finalizer():
        child.stop()

    request.addfinalizer(finalizer)

    return app


def test_return_value(app, watcher):
    event = Event.create('foo')
    event.notify = True

    remote_event = remote(event, 'child')
    remote_event.notify = True

    value = app.fire(remote_event)
    assert watcher.wait('remote_value_changed')
    assert value.value == 'Hello World!'



#!/usr/bin/env python
from http.client import HTTPConnection

from circuits.web import Controller


class Root(Controller):
    def index(self):
        return 'Hello World!'


def test(webapp):
    connection = HTTPConnection(webapp.server.host, webapp.server.port)
    connection.auto_open = False
    connection.connect()

    for _i in range(2):
        connection.request('GET', '/')
        response = connection.getresponse()
        assert response.status == 200
        assert response.reason == 'OK'
        s = response.read()
        assert s == b'Hello World!'

    connection.close()



"""py.test config"""

import sys
import threading
from collections import deque
from time import sleep

import pytest

from circuits import BaseComponent, Debugger, Manager, handler
from circuits.core.manager import TIMEOUT


class Watcher(BaseComponent):
    def init(self):
        self._lock = threading.Lock()
        self.events = deque()

    @handler(channel='*', priority=999.9)
    def _on_event(self, event, *args, **kwargs):
        with self._lock:
            self.events.append(event)

    def wait(self, name, channel=None, timeout=6.0):
        for _i in range(int(timeout / TIMEOUT)):
            if channel is None:
                with self._lock:
                    for event in self.events:
                        if event.name == name:
                            return True
            else:
                with self._lock:
                    for event in self.events:
                        if event.name == name and channel in event.channels:
                            return True

            sleep(TIMEOUT)
        return None


class Flag:
    status = False


class WaitEvent:
    def __init__(self, manager, name, channel=None, timeout=6.0):
        if channel is None:
            channel = getattr(manager, 'channel', None)

        self.timeout = timeout
        self.manager = manager

        flag = Flag()

        @handler(name, channel=channel)
        def on_event(self, *args, **kwargs):
            flag.status = True

        self.handler = self.manager.addHandler(on_event)
        self.flag = flag

    def wait(self):
        try:
            for _i in range(int(self.timeout / TIMEOUT)):
                if self.flag.status:
                    return True
                sleep(TIMEOUT)
        finally:
            self.manager.removeHandler(self.handler)


@pytest.fixture(scope='session')
def manager(request):
    manager = Manager()

    def finalizer():
        manager.stop()

    request.addfinalizer(finalizer)

    waiter = WaitEvent(manager, 'started')
    manager.start()
    assert waiter.wait()

    if request.config.option.verbose:
        Debugger().register(manager)

    return manager


@pytest.fixture()
def watcher(request, manager):
    watcher = Watcher().register(manager)

    def finalizer():
        waiter = WaitEvent(manager, 'unregistered')
        watcher.unregister()
        waiter.wait()

    request.addfinalizer(finalizer)

    return watcher


for key, value in {'WaitEvent': WaitEvent, 'PLATFORM': sys.platform, 'PYVER': sys.version_info[:3]}.items():
    setattr(pytest, key, value)



"""This module defines the basic event class and common events."""

from inspect import ismethod
from traceback import format_tb


class Event:
    channels = ()
    'The channels this message is sent to.'

    parent = None
    notify = False
    success = False
    failure = False
    complete = False
    alert_done = False
    waitingHandlers = 0

    @classmethod
    def create(cls, _name, *args, **kwargs):
        return type(cls)(_name, (cls,), {})(*args, **kwargs)

    def child(self, name, *args, **kwargs):
        e = Event.create(f'{self.name:s}_{name:s}', *args, **kwargs)
        e.parent = self
        return e

    def __init__(self, *args, **kwargs):
        """
        An event is a message send to one or more channels.
        It is eventually dispatched to all components
        that have handlers for one of the channels and the event type.

        All normal arguments and keyword arguments passed to the constructor
        of an event are passed on to the handler. When declaring a
        handler, its argument list must therefore match the arguments
        used for creating the event.

        Every event has a :attr:`name` attribute that is used for matching
        the event with the handlers.

        :cvar channels: an optional attribute that may be set before firing
            the event. If defined (usually as a class variable), the attribute
            specifies the channels that the event should be delivered
            to as a tuple. This overrides the default behavior
            of sending the event to the firing component's channel.

            When an event is fired, the value in this attribute
            is replaced for the instance with the channels that
            the event is actually sent to. This information may be used
            e.g. when the event is passed as a parameter to a handler.

        :ivar value: this is a :class:`circuits.core.values.Value`
            object that holds the results returned by the handlers invoked
            for the event.

        :var success: if this optional attribute is set to
            ``True``, an associated event ``success`` (original name
            with "_success" appended) will automatically be fired when all
            handlers for the event have been invoked successfully.

        :var success_channels: the success event is, by default, delivered
            to same channels as the successfully dispatched event itself.
            This may be overridden by specifying an alternative list of
            destinations using this attribute.

        :var complete: if this optional attribute is set to
            ``True``, an associated event ``complete`` (original name
            with "_complete" appended) will automatically be fired when all
            handlers for the event and all events fired by these handlers
            (recursively) have been invoked successfully.

        :var complete_channels: the complete event is, by default, delivered
            to same channels as the initially dispatched event itself.
            This may be overridden by specifying an alternative list of
            destinations using this attribute.
        """
        self.args = list(args)
        self.kwargs = kwargs

        self.uid = None
        self.value = None
        self.handler = None
        self.stopped = False
        self.cancelled = False
        if not hasattr(self, 'name'):
            self.name = self.__class__.__name__

    def __getstate__(self):
        odict = self.__dict__.copy()
        del odict['handler']
        return odict

    def __setstate__(self, dict):
        self.__dict__.update(dict)

    def __le__(self, other):
        return False

    def __gt__(self, other):
        return False

    def __repr__(self):
        """x.__repr__() <==> repr(x)"""
        if len(self.channels) > 1:
            channels = repr(self.channels)
        elif len(self.channels) == 1:
            channels = str(self.channels[0])
        else:
            channels = ''

        data = '%s %s' % (
            ', '.join(repr(arg) for arg in self.args),
            ', '.join(f'{k}={v!r}' for k, v in self.kwargs.items()),
        )

        return f'<{self.name}[{channels}] ({data})>'

    def __getitem__(self, x):
        """
        x.__getitem__(y) <==> x[y]

        Get and return data from the event object requested by "x".
        If an int is passed to x, the requested argument from self.args
        is returned index by x. If a str is passed to x, the requested
        keyword argument from self.kwargs is returned keyed by x.
        Otherwise a TypeError is raised as nothing else is valid.
        """
        if isinstance(x, int):
            return self.args[x]
        if isinstance(x, str):
            return self.kwargs[x]
        raise TypeError('Expected int or str, got %r' % type(x))

    def __setitem__(self, i, y):
        """
        x.__setitem__(i, y) <==> x[i] = y

        Modify the data in the event object requested by "x".
        If i is an int, the ith requested argument from self.args
        shall be changed to y. If i is a str, the requested value
        keyed by i from self.kwargs, shall by changed to y.
        Otherwise a TypeError is raised as nothing else is valid.
        """
        if isinstance(i, int):
            self.args[i] = y
        elif isinstance(i, str):
            self.kwargs[i] = y
        else:
            raise TypeError('Expected int or str, got %r' % type(i))

    def cancel(self):
        """Cancel the event from being processed (if not already)"""
        self.cancelled = True

    def stop(self):
        """Stop further processing of this event"""
        self.stopped = True


class exception(Event):
    """
    exception Event

    This event is sent for any exceptions that occur during the execution
    of an event Handler that is not SystemExit or KeyboardInterrupt.

    :param type: type of exception
    :type  type: type

    :param value: exception object
    :type  value: exceptions.Exceptions

    :param traceback: traceback of exception
    :type  traceback: traceback

    :param handler: handler that raised the exception
    :type  handler: @handler(<method>)

    :param fevent: event that failed
    :type  fevent: event
    """

    def __init__(self, type, value, traceback, handler=None, fevent=None):
        super().__init__(type, value, self.format_traceback(traceback), handler=handler, fevent=fevent)

    def format_traceback(self, traceback):
        return format_tb(traceback)


class started(Event):
    """
    started Event

    This Event is sent when a Component or Manager has started running.

    :param manager: The component or manager that was started
    :type  manager: Component or Manager
    """

    def __init__(self, manager):
        super().__init__(manager)


class stopped(Event):
    """
    stopped Event

    This Event is sent when a Component or Manager has stopped running.

    :param manager: The component or manager that has stopped
    :type  manager: Component or Manager
    """

    def __init__(self, manager):
        super().__init__(manager)


class signal(Event):
    """
    signal Event

    This Event is sent when a Component receives a signal.

    :param signo: The signal number received.
    :type  int:   An int value for the signal

    :param stack:  The interrupted stack frame.
    :type  object: A stack frame
    """

    def __init__(self, signo, stack):
        super().__init__(signo, stack)


class registered(Event):
    """
    registered Event

    This Event is sent when a Component has registered with another Component
    or Manager. This Event is only sent if the Component or Manager being
    registered which is not itself.

    :param component: The Component being registered
    :type  component: Component

    :param manager: The Component or Manager being registered with
    :type  manager: Component or Manager
    """

    def __init__(self, component, manager):
        super().__init__(component, manager)


class unregistered(Event):
    """
    unregistered Event

    This Event is sent when a Component has been unregistered from its
    Component or Manager.
    """


class generate_events(Event):
    """
    generate_events Event

    This Event is sent by the circuits core. All components that generate
    timed events or events from external sources (e.g. data becoming
    available) should fire any pending events in their "generate_events"
    handler.

    The handler must either call :meth:`~stop` (*preventing other handlers
    from being called in the same iteration*)
    or must invoke :meth:`~.reduce_time_left` with parameter 0.

    :param max_wait: maximum time available for generating events.
    :type  time_left: float

    Components that actually consume time waiting for events to be generated,
    thus suspending normal execution, must provide a method ``resume``
    that interrupts waiting for events.
    """

    def __init__(self, lock, max_wait):
        super().__init__()

        self._time_left = max_wait
        self._lock = lock

    @property
    def time_left(self):
        """
        The time left for generating events. A value less than 0
        indicates unlimited time. You should have only
        one component in your system (usually a poller component) that
        spends up to "time left" until it generates an event.
        """
        return self._time_left

    def reduce_time_left(self, time_left):
        """
        Update the time left for generating events. This is typically
        used by event generators that currently don't want to generate
        an event but know that they will within a certain time. By
        reducing the time left, they make sure that they are reinvoked
        when the time for generating the event has come (at the latest).

        This method can only be used to reduce the time left. If the
        parameter is larger than the current value of time left, it is
        ignored.

        If the time left is reduced to 0 and the event is currently
        being handled, the handler's *resume* method is invoked.
        """
        with self._lock:
            if time_left >= 0 and (self._time_left < 0 or self._time_left > time_left):
                self._time_left = time_left
                if self._time_left == 0 and self.handler is not None:
                    m = getattr(
                        getattr(
                            self.handler,
                            'im_self',
                            self.handler.__self__,
                        ),
                        'resume',
                        None,
                    )
                    if m is not None and ismethod(m):
                        m()

    @property
    def lock(self):
        return self._lock



"""
test_utils

...
"""

from circuits import Event
from circuits.core import Value
from circuits.node.utils import dump_event, dump_value, load_event, load_value


class test(Event):
    """test Event"""


def test_events():
    event = test(1, 2, 3, foo='bar')
    event.success = True
    event.failure = False
    event.test_meta = 1

    id = 1
    s = dump_event(event, id)
    event, id = load_event(s)

    assert hasattr(event, 'args')
    assert hasattr(event, 'kwargs')
    assert hasattr(event, 'success')
    assert hasattr(event, 'failure')
    assert hasattr(event, 'channels')
    assert hasattr(event, 'notify')
    assert hasattr(event, 'test_meta')


def test_values():
    event = Event()
    event.test_meta = 1

    value = Value(event=event)
    value.value = 'foo'
    value.errors = False
    value.node_call_id = 1

    x, id, errors, meta = load_value(dump_value(value))

    assert value.value == x
    assert id == 1
    assert not errors
    assert meta['test_meta'] == event.test_meta



#!/usr/bin/env python
import contextlib
import os.path
import select
from socket import AF_INET, AF_INET6, EAI_NODATA, EAI_NONAME, SOCK_STREAM, has_ipv6, socket
from ssl import CERT_NONE, PROTOCOL_TLS_CLIENT, SSLContext

import pytest

from circuits import Debugger, Manager
from circuits.core.pollers import EPoll, KQueue, Poll, Select
from circuits.net.events import close, connect, write
from circuits.net.sockets import TCP6Client, TCP6Server, TCPClient, TCPServer
from tests.conftest import WaitEvent

from .client import Client
from .server import Server


CERT_FILE = os.path.join(os.path.dirname(__file__), 'cert.pem')


class _TestClient:
    def __init__(self, ipv6=False):
        self._sockname = None

        self.sock = socket(
            AF_INET6 if ipv6 else AF_INET,
            SOCK_STREAM,
        )

        ctx = SSLContext(PROTOCOL_TLS_CLIENT)
        ctx.check_hostname = False
        ctx.verify_mode = CERT_NONE
        self.ssock = ctx.wrap_socket(self.sock)

    @property
    def sockname(self):
        return self._sockname

    def connect(self, host, port):
        self.ssock.connect_ex((host, port))
        self._sockname = self.ssock.getsockname()

    def send(self, data):
        self.ssock.send(data)

    def recv(self, buflen=4069):
        return self.ssock.recv(buflen)

    def disconnect(self):
        with contextlib.suppress(OSError):
            self.ssock.shutdown(2)

        with contextlib.suppress(OSError):
            self.ssock.close()


@pytest.fixture()
def client(request, ipv6):
    client = _TestClient(ipv6=ipv6)

    def finalizer():
        client.disconnect()

    request.addfinalizer(finalizer)

    return client


def wait_host(server):
    def checker(obj, attr):
        return all(getattr(obj, a) for a in attr)

    assert pytest.wait_for(server, ('host', 'port'), checker)


def pytest_generate_tests(metafunc):
    ipv6 = [False]
    if has_ipv6:
        ipv6.append(True)

    for ipv6 in ipv6:
        poller = [(Select, ipv6)]

        if hasattr(select, 'poll'):
            poller.append((Poll, ipv6))

        if hasattr(select, 'epoll'):
            poller.append((EPoll, ipv6))

        if hasattr(select, 'kqueue'):
            poller.append((KQueue, ipv6))

    metafunc.parametrize('Poller,ipv6', poller)


def test_tcp_basic(Poller, ipv6):
    m = Manager() + Poller()

    if ipv6:
        tcp_server = TCP6Server(('::1', 0))
        tcp_client = TCP6Client()
    else:
        tcp_server = TCPServer(0)
        tcp_client = TCPClient()
    server = Server() + tcp_server
    client = Client() + tcp_client

    server.register(m)
    client.register(m)

    m.start()

    try:
        assert pytest.wait_for(client, 'ready')
        assert pytest.wait_for(server, 'ready')
        wait_host(server)

        client.fire(connect(server.host, server.port))
        assert pytest.wait_for(client, 'connected')
        assert pytest.wait_for(server, 'connected')
        assert pytest.wait_for(client, 'data', b'Ready')

        client.fire(write(b'foo'))
        assert pytest.wait_for(server, 'data', b'foo')
        assert pytest.wait_for(client, 'data', b'foo')

        client.fire(close())
        assert pytest.wait_for(client, 'disconnected')
        assert pytest.wait_for(server, 'disconnected')

        server.fire(close())
        assert pytest.wait_for(server, 'closed')
    finally:
        m.stop()


def test_tcps_basic(manager, watcher, client, Poller, ipv6):
    poller = Poller().register(manager)

    if ipv6:
        tcp_server = TCP6Server(('::1', 0), secure=True, certfile=CERT_FILE)
    else:
        tcp_server = TCPServer(0, secure=True, certfile=CERT_FILE)

    server = Server() + tcp_server

    server.register(manager)

    try:
        watcher.wait('ready', 'server')

        client.connect(server.host, server.port)
        assert watcher.wait('connect', 'server')
        assert client.recv() == b'Ready'

        client.send(b'foo')
        assert watcher.wait('read', 'server')
        assert client.recv() == b'foo'

        client.disconnect()
        assert watcher.wait('disconnect', 'server')

        server.fire(close())
        assert watcher.wait('closed', 'server')
    finally:
        poller.unregister()
        server.unregister()


def test_tcp_reconnect(Poller, ipv6):
    # TODO: Apparently this doesn't work on Windows either?
    # TODO: UPDATE: Apparently Broken on Windows + Python 3.2
    # TODO: Need to look into this. Find out why...

    if pytest.PLATFORM == 'win32' and pytest.PYVER[:2] >= (3, 2):
        pytest.skip('Broken on Windows on Python 3.2')

    m = Manager() + Poller()

    if ipv6:
        tcp_server = TCP6Server(('::1', 0))
        tcp_client = TCP6Client()
    else:
        tcp_server = TCPServer(0)
        tcp_client = TCPClient()
    server = Server() + tcp_server
    client = Client() + tcp_client

    server.register(m)
    client.register(m)

    m.start()

    try:
        assert pytest.wait_for(client, 'ready')
        assert pytest.wait_for(server, 'ready')
        wait_host(server)

        # 1st connect
        client.fire(connect(server.host, server.port))
        assert pytest.wait_for(client, 'connected')
        assert pytest.wait_for(server, 'connected')
        assert pytest.wait_for(client, 'data', b'Ready')

        client.fire(write(b'foo'))
        assert pytest.wait_for(server, 'data', b'foo')

        # disconnect
        client.fire(close())
        assert pytest.wait_for(client, 'disconnected')

        # 2nd reconnect
        client.fire(connect(server.host, server.port))
        assert pytest.wait_for(client, 'connected')
        assert pytest.wait_for(server, 'connected')
        assert pytest.wait_for(client, 'data', b'Ready')

        client.fire(write(b'foo'))
        assert pytest.wait_for(server, 'data', b'foo')

        client.fire(close())
        assert pytest.wait_for(client, 'disconnected')
        assert pytest.wait_for(server, 'disconnected')

        server.fire(close())
        assert pytest.wait_for(server, 'closed')
    finally:
        m.stop()


def test_tcp_connect_closed_port(Poller, ipv6):
    if pytest.PLATFORM == 'win32':
        pytest.skip('Broken on Windows')

    m = Manager() + Poller() + Debugger()

    if ipv6:
        tcp_server = TCP6Server(('::1', 0))
        tcp_client = TCP6Client(connect_timeout=1)
    else:
        tcp_server = TCPServer(0)
        tcp_client = TCPClient(connect_timeout=1)
    server = Server() + tcp_server
    client = Client() + tcp_client

    server.register(m)
    client.register(m)

    m.start()

    try:
        assert pytest.wait_for(client, 'ready')
        assert pytest.wait_for(server, 'ready')
        wait_host(server)

        host, port = server.host, server.port
        tcp_server._sock.close()

        # 1st connect
        client.fire(connect(host, port))
        waiter = WaitEvent(m, 'unreachable', channel='client')
        assert waiter.wait()
    finally:
        server.unregister()
        client.unregister()
        m.stop()


def test_tcp_bind(Poller, ipv6):
    m = Manager() + Poller()

    if ipv6:
        sock = socket(AF_INET6, SOCK_STREAM)
        sock.bind(('::1', 0))
        sock.listen(5)
        _, _bind_port, _, _ = sock.getsockname()
        sock.close()
        server = Server() + TCP6Server(('::1', 0))
        client = Client() + TCP6Client()
    else:
        sock = socket(AF_INET, SOCK_STREAM)
        sock.bind(('', 0))
        sock.listen(5)
        _, _bind_port = sock.getsockname()
        sock.close()
        server = Server() + TCPServer(0)
        client = Client() + TCPClient()

    server.register(m)
    client.register(m)

    m.start()

    try:
        assert pytest.wait_for(client, 'ready')
        assert pytest.wait_for(server, 'ready')
        wait_host(server)

        client.fire(connect(server.host, server.port))
        assert pytest.wait_for(client, 'connected')
        assert pytest.wait_for(server, 'connected')
        assert pytest.wait_for(client, 'data', b'Ready')

        # assert server.client[1] == bind_port

        client.fire(write(b'foo'))
        assert pytest.wait_for(server, 'data', b'foo')

        client.fire(close())
        assert pytest.wait_for(client, 'disconnected')
        assert pytest.wait_for(server, 'disconnected')

        server.fire(close())
        assert pytest.wait_for(server, 'closed')
    finally:
        m.stop()


def test_tcp_lookup_failure(manager, watcher, Poller, ipv6):
    poller = Poller().register(manager)

    tcp_client = TCP6Client() if ipv6 else TCPClient()

    client = Client() + tcp_client
    client.register(manager)

    try:
        assert watcher.wait('ready', 'client')

        client.fire(connect('foo.bar.baz', 1234))
        assert watcher.wait('error', 'client')

        if pytest.PLATFORM == 'win32':
            assert client.error.errno == 11004
        else:
            assert client.error.errno in (EAI_NODATA, EAI_NONAME)
    finally:
        poller.unregister()
        client.unregister()



#!/usr/bin/env python
from circuits.web import Controller, expose

from .helpers import urlopen


class Root(Controller):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self += Hello()
        self += World()

    def index(self):
        return 'index'

    def hello1(self):
        return 'hello1'

    @expose('hello2')
    def hello2(self):
        return 'hello2'

    def query(req, test):
        return 'query %s' % test


class Hello(Controller):
    channel = '/hello'

    def index(self):
        return 'hello index'

    def test(self):
        return 'hello test'

    def query(req, test):
        return 'hello query %s' % test


class World(Controller):
    channel = '/world'

    def index(self):
        return 'world index'

    def test(self):
        return 'world test'


def test_simple(webapp):
    url = '%s/hello1' % webapp.server.http.base
    f = urlopen(url)
    s = f.read()
    assert s == b'hello1'


def test_expose(webapp):
    url = '%s/hello2' % webapp.server.http.base
    f = urlopen(url)
    s = f.read()
    assert s == b'hello2'


def test_index(webapp):
    f = urlopen(webapp.server.http.base)
    s = f.read()
    assert s == b'index'


def test_controller_index(webapp):
    url = '%s/hello/' % webapp.server.http.base
    f = urlopen(url)
    s = f.read()
    assert s == b'hello index'

    url = '%s/world/' % webapp.server.http.base
    f = urlopen(url)
    s = f.read()
    assert s == b'world index'


def test_controller_expose(webapp):
    url = '%s/hello/test' % webapp.server.http.base
    f = urlopen(url)
    s = f.read()
    assert s == b'hello test'

    url = '%s/world/test' % webapp.server.http.base
    f = urlopen(url)
    s = f.read()
    assert s == b'world test'


def test_query(webapp):
    url = '%s/query?test=1' % webapp.server.http.base
    f = urlopen(url)
    s = f.read()
    assert s == b'query 1'

    url = '%s/hello/query?test=2' % webapp.server.http.base
    f = urlopen(url)
    s = f.read()
    assert s == b'hello query 2'



#!/usr/bin/env python
from circuits.web import Server
from circuits.web.controllers import BaseController, expose


class Root(BaseController):
    @expose('index')
    def index(self):
        """
        Index Request Handler

        BaseController(s) do not expose methods as request handlers.
        Request Handlers _must_ be exposed explicitly using the ``@expose``
        decorator.
        """
        return 'Hello World!'


app = Server(('0.0.0.0', 8000))
Root().register(app)
app.run()



#!/usr/bin/env python
from circuits import BaseComponent, Event, handler


class test(Event):
    """test Event"""


class App(BaseComponent):
    @handler('test')
    def _on_test(self, event):
        try:
            return 'Hello World!'
        finally:
            event.stop()

    def _on_test2(self):
        pass  # Never reached


def test_main():
    app = App()
    while len(app):
        app.flush()
    x = app.fire(test())
    app.flush()
    assert x.value == 'Hello World!'



#!/usr/bin/env python
"""
curl -i http://localhost:8000/ -H 'Content-Type: application/json' \
-d '{"id": "test", "method": "test", "params": {"foo": "bar"}}'
"""

from circuits import Component
from circuits.web import JSONRPC, Logger, Server


class Test(Component):
    def foo(self, a, b, c):
        return a, b, c


app = Server(('0.0.0.0', 8000))
Logger().register(app)
JSONRPC().register(app)
Test().register(app)
app.run()



#!/usr/bin/env python
import pytest

from circuits.web import Server
from circuits.web.wsgi import Gateway

from .helpers import urlopen


def hello(environ, start_response):
    status = '200 OK'
    response_headers = [('Content-type', 'text/plain')]
    start_response(status, response_headers)
    return 'Hello World!'


def foobar(environ, start_response):
    status = '200 OK'
    response_headers = [('Content-type', 'text/plain')]
    start_response(status, response_headers)
    return 'FooBar!'


@pytest.fixture()
def apps(request):
    return {
        '/': hello,
        '/foobar': foobar,
    }


def test(apps):
    server = Server(0)
    Gateway(apps).register(server)

    waiter = pytest.WaitEvent(server, 'ready')
    server.start()
    waiter.wait()

    f = urlopen(server.http.base)
    s = f.read()
    assert s == b'Hello World!'
    assert f.headers.get('Transfer-Encoding') != 'chunked'

    f = urlopen(f'{server.http.base:s}/foobar/')
    s = f.read()
    assert s == b'FooBar!'
    assert f.headers.get('Transfer-Encoding') != 'chunked'

    server.stop()



"""Internet Relay Chat message"""

from .utils import parsemsg


class Error(Exception):
    """Error Exception"""


class Message:
    def __init__(self, command, *args, **kwargs):
        self.command = command
        self.prefix = str(kwargs['prefix']) if 'prefix' in kwargs else None

        self.encoding = kwargs.get('encoding', 'utf-8')
        self.add_nick = kwargs.get('add_nick', False)
        self.args = [arg if isinstance(arg, str) else arg.decode(self.encoding) for arg in args if arg is not None]
        self._check_args()

    def _check_args(self):
        if any(type(arg)(' ') in arg in arg for arg in self.args[:-1] if isinstance(arg, str)):
            raise Error('Space can only appear in the very last arg')
        if any(type(arg)('\n') in arg for arg in self.args if isinstance(arg, str)):
            raise Error('No newline allowed')

    @staticmethod
    def from_string(s):
        if len(s) > 512:
            raise Error('Message must not be longer than 512 characters')

        prefix, command, args = parsemsg(s)

        return Message(command, *args, prefix=prefix)

    def __bytes__(self):
        return str(self).encode(self.encoding)

    def __str__(self):
        self._check_args()
        args = self.args[:]

        if args and ' ' in args[-1] and not args[-1].startswith(':'):
            args[-1] = f':{args[-1]}'

        return '{prefix}{command} {args}\r\n'.format(
            prefix=(f':{self.prefix} ' if self.prefix is not None else ''),
            command=str(self.command),
            args=' '.join(args),
        )

    def __repr__(self):
        return repr(str(self)[:-2])

    def __eq__(self, other):
        return (
            isinstance(other, Message)
            and self.prefix == other.prefix
            and self.command == other.command
            and self.args == other.args
        )



#!/usr/bin/env python
"""
Main

circutis.web Web Server and Testing Tool.
"""

import os
from argparse import ArgumentParser
from hashlib import md5
from sys import stderr
from wsgiref.simple_server import make_server
from wsgiref.validate import validator

import circuits
from circuits import Component, Debugger, Manager, handler
from circuits.core.pollers import Select
from circuits.tools import graph, inspect
from circuits.web import BaseServer, Controller, Logger, Server, Static
from circuits.web.tools import check_auth, digest_auth
from circuits.web.wsgi import Application


try:
    import hotshot
    import hotshot.stats
except ImportError:
    hostshot = None


try:
    from circuits.core.pollers import Poll
except ImportError:
    Poll = None  # NOQA

try:
    from circuits.core.pollers import EPoll
except ImportError:
    EPoll = None  # NOQA


def parse_options():
    parser = ArgumentParser()

    parser.add_argument(
        '-b',
        '--bind',
        action='store',
        default='0.0.0.0:8000',
        help='Bind to address:[port]',
    )

    parser.add_argument(
        '-l',
        '--logging',
        action='store_true',
        default=False,
        help='Enable logging of requests',
    )

    parser.add_argument(
        '-p',
        '--passwd',
        action='store',
        default=None,
        help='Location to passwd file for Digest Auth',
    )

    parser.add_argument(
        '-j',
        '--jobs',
        action='store',
        type=int,
        default=0,
        help='Specify number of jobs/processes to start',
    )

    parser.add_argument(
        '--poller',
        action='store',
        default='select',
        help='Specify type of poller to use',
    )

    parser.add_argument(
        '--server',
        action='store',
        default='server',
        help='Specify server to use',
    )

    parser.add_argument(
        '--profile',
        action='store_true',
        default=False,
        help='Enable execution profiling support',
    )

    parser.add_argument(
        '--debug',
        action='store_true',
        default=False,
        help='Enable debug mode',
    )

    parser.add_argument(
        '--validate',
        action='store_true',
        default=False,
        help='Enable WSGI validation mode',
    )

    parser.add_argument('-v', '--version', action='version', version=f'%(prog)s v{circuits.__version__}')

    parser.add_argument('docroot', nargs='?', default=os.getcwd())

    return parser.parse_args()


class Authentication(Component):
    channel = 'web'

    realm = 'Secure Area'
    users = {'admin': md5(b'admin').hexdigest()}

    def __init__(self, channel=channel, realm=None, passwd=None):
        super().__init__(self, channel=channel)

        if realm is not None:
            self.realm = realm

        if passwd is not None:
            with open(passwd) as f:
                lines = (line.strip() for line in f)
                self.users = dict(line.split(':', 1) for line in lines)

    @handler('request', priority=10)
    def request(self, event, request, response):
        if not check_auth(request, response, self.realm, self.users):
            event.stop()
            return digest_auth(request, response, self.realm, self.users)
        return None


class HelloWorld(Component):
    channel = 'web'

    def request(self, request, response):
        return 'Hello World!'


class Root(Controller):
    def hello(self):
        return 'Hello World!'


def select_poller(poller):
    if poller == 'poll':
        if Poll is None:
            stderr.write(
                'No poll support available - defaulting to Select...',
            )
            Poller = Select
        else:
            Poller = Poll
    elif poller == 'epoll':
        if EPoll is None:
            stderr.write(
                'No epoll support available - defaulting to Select...',
            )
            Poller = Select
        else:
            Poller = EPoll
    else:
        Poller = Select

    return Poller


def parse_bind(bind):
    if ':' in bind:
        address, port = bind.split(':')
        port = int(port)
    else:
        address, port = bind, 8000

    return (address, port)


def main():
    opts = parse_options()

    bind = parse_bind(opts.bind)

    if opts.validate:
        application = Application() + Root()
        app = validator(application)

        httpd = make_server(bind[0], bind[1], app)
        httpd.serve_forever()

        raise SystemExit(0)

    manager = Manager()

    opts.debug and Debugger().register(manager)

    Poller = select_poller(opts.poller.lower())
    Poller().register(manager)

    if opts.server.lower() == 'base':
        BaseServer(bind).register(manager)
        HelloWorld().register(manager)
    else:
        Server(bind).register(manager)
        Root().register(manager)

    Static(docroot=opts.docroot, dirlisting=True).register(manager)

    opts.passwd and Authentication(passwd=opts.passwd).register(manager)

    opts.logging and Logger().register(manager)

    if opts.profile and hotshot:
        profiler = hotshot.Profile('.profile')
        profiler.start()

    if opts.debug:
        print(graph(manager, name='circuits.web'))
        print()
        print(inspect(manager))

    for _i in range(opts.jobs):
        manager.start(process=True)

    manager.run()

    if opts.profile and hotshot:
        profiler.stop()
        profiler.close()

        stats = hotshot.stats.load('.profile')
        stats.strip_dirs()
        stats.sort_stats('time', 'calls')
        stats.print_stats(20)


if __name__ == '__main__':
    main()



"""
Dispatcher

This module implements a basic URL to Channel dispatcher.
This is the default dispatcher used by circuits.web
"""

from circuits import BaseComponent, Event, handler
from circuits.web.controllers import BaseController
from circuits.web.events import response
from circuits.web.processors import process
from circuits.web.utils import parse_qs


def resolve_path(paths, parts):
    def rebuild_path(url_parts):
        return '/%s' % '/'.join(url_parts)

    left_over = []

    while parts:
        if rebuild_path(parts) in paths:
            yield rebuild_path(parts), left_over
        left_over.insert(0, parts.pop())

    if '/' in paths:
        yield '/', left_over


def resolve_methods(parts):
    if parts:
        method = parts[0]
        vpath = parts[1:]
        yield method, vpath

    yield 'index', parts


def find_handlers(req, paths):
    def get_handlers(path, method):
        component = paths[path]
        return component._handlers.get(method, None)

    def accepts_vpath(handlers, vpath):
        args_no = len(vpath)
        return all(
            len(h.args) == args_no or h.varargs or (h.defaults is not None and args_no <= len(h.defaults)) for h in handlers
        )

    # Split /hello/world to ['hello', 'world']
    starting_parts = [x for x in req.path.strip('/').split('/') if x]

    for path, parts in resolve_path(paths, starting_parts):
        handlers = get_handlers(path, req.method)
        if handlers:
            return handlers, req.method, path, parts

        for method, vpath in resolve_methods(parts):
            handlers = get_handlers(path, method)
            if handlers and (not vpath or accepts_vpath(handlers, vpath)):
                req.index = method == 'index'
                return handlers, method, path, vpath
            method, vpath = 'index', [method] + vpath
            handlers = get_handlers(path, method)
            if handlers and (not vpath or accepts_vpath(handlers, vpath)):
                req.index = True
                return handlers, method, path, vpath

    return [], None, None, None


class Dispatcher(BaseComponent):
    channel = 'web'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.paths = {}

    @handler('registered', channel='*')
    def _on_registered(self, component, manager):
        if isinstance(component, BaseController) and component.channel not in self.paths:
            self.paths[component.channel] = component

    @handler('unregistered', channel='*')
    def _on_unregistered(self, component, manager):
        if isinstance(component, BaseController) and component.channel in self.paths:
            del self.paths[component.channel]

    @handler('request', priority=0.1)
    def _on_request(self, event, req, res, peer_cert=None):
        if peer_cert:
            event.peer_cert = peer_cert

        _handlers, name, channel, vpath = find_handlers(req, self.paths)

        if name is not None and channel is not None:
            event.kwargs = parse_qs(req.qs)
            process(req, event.kwargs)

            if vpath:
                event.args += tuple(vpath)

            return self.fire(
                Event.create(
                    name,
                    *event.args,
                    **event.kwargs,
                ),
                channel,
            )
        return None

    @handler('request_value_changed')
    def _on_request_value_changed(self, value):
        if value.handled:
            return

        _req, res = value.event.args[:2]
        if value.result and not value.errors:
            res.body = value.value
            self.fire(response(res))
        elif value.promise:
            value.event.notify = True
        else:
            # Errors are handled by the ``HTTP`` Protocol Component
            return
